"""
Use a postgresql server for the master job cache. This helps the job cache to
cope with scale.

.. note::
    There are three PostgreSQL returners.  Any can function as an external
    :ref:`master job cache <external-job-cache>`. but each has different
    features.  SaltStack recommends
    :mod:`returners.pgjsonb <salt.returners.pgjsonb>` if you are working with
    a version of PostgreSQL that has the appropriate native binary JSON types.
    Otherwise, review
    :mod:`returners.postgres <salt.returners.postgres>` and
    :mod:`returners.postgres_local_cache <salt.returners.postgres_local_cache>`
    to see which module best suits your particular needs.

:maintainer:    gjredelinghuys@gmail.com
:maturity:      Stable
:depends:       psycopg2
:platform:      all

To enable this returner the minion will need the psycopg2 installed and
the following values configured in the master config:

.. code-block:: yaml

    master_job_cache: postgres_local_cache
    master_job_cache.postgres.host: 'salt'
    master_job_cache.postgres.user: 'salt'
    master_job_cache.postgres.passwd: 'salt'
    master_job_cache.postgres.db: 'salt'
    master_job_cache.postgres.port: 5432

Running the following command as the postgres user should create the database
correctly:

.. code-block:: sql

    psql << EOF
    CREATE ROLE salt WITH PASSWORD 'salt';
    CREATE DATABASE salt WITH OWNER salt;
    EOF

In case the postgres database is a remote host, you'll need this command also:

.. code-block:: sql

   ALTER ROLE salt WITH LOGIN;

and then:

.. code-block:: sql

    psql -h localhost -U salt << EOF
    --
    -- Table structure for table 'jids'
    --

    DROP TABLE IF EXISTS jids;
    CREATE TABLE jids (
      jid   varchar(20) PRIMARY KEY,
      started TIMESTAMP WITH TIME ZONE DEFAULT now(),
      tgt_type text NOT NULL,
      cmd text NOT NULL,
      tgt text NOT NULL,
      kwargs text NOT NULL,
      ret text NOT NULL,
      username text NOT NULL,
      arg text NOT NULL,
      fun text NOT NULL
    );

    --
    -- Table structure for table 'salt_returns'
    --
    -- note that 'success' must not have NOT NULL constraint, since
    -- some functions don't provide it.

    DROP TABLE IF EXISTS salt_returns;
    CREATE TABLE salt_returns (
      added     TIMESTAMP WITH TIME ZONE DEFAULT now(),
      fun       text NOT NULL,
      jid       varchar(20) NOT NULL,
      return    text NOT NULL,
      id        text NOT NULL,
      success   boolean
    );
    CREATE INDEX ON salt_returns (added);
    CREATE INDEX ON salt_returns (id);
    CREATE INDEX ON salt_returns (jid);
    CREATE INDEX ON salt_returns (fun);

    DROP TABLE IF EXISTS salt_events;
    CREATE TABLE salt_events (
      id SERIAL,
      tag text NOT NULL,
      data text NOT NULL,
      alter_time TIMESTAMP WITH TIME ZONE DEFAULT now(),
      master_id text NOT NULL
    );
    CREATE INDEX ON salt_events (tag);
    CREATE INDEX ON salt_events (data);
    CREATE INDEX ON salt_events (id);
    CREATE INDEX ON salt_events (master_id);
    EOF

Required python modules: psycopg2
"""


import logging
import re
import sys

import salt.utils.jid
import salt.utils.json

try:
    import psycopg2

    HAS_POSTGRES = True
except ImportError:
    HAS_POSTGRES = False

log = logging.getLogger(__name__)

__virtualname__ = "postgres_local_cache"


def __virtual__():
    if not HAS_POSTGRES:
        return (False, "Could not import psycopg2; postges_local_cache disabled")
    return __virtualname__


def _get_conn():
    """
    Return a postgres connection.
    """
    try:
        conn = psycopg2.connect(
            host=__opts__["master_job_cache.postgres.host"],
            user=__opts__["master_job_cache.postgres.user"],
            password=__opts__["master_job_cache.postgres.passwd"],
            database=__opts__["master_job_cache.postgres.db"],
            port=__opts__["master_job_cache.postgres.port"],
        )
    except psycopg2.OperationalError:
        log.error("Could not connect to SQL server: %s", sys.exc_info()[0])
        return None
    return conn


def _close_conn(conn):
    """
    Close the postgres connection.
    """
    conn.commit()
    conn.close()


def _format_job_instance(job):
    """
    Format the job instance correctly
    """
    ret = {
        "Function": job.get("fun", "unknown-function"),
        "Arguments": salt.utils.json.loads(job.get("arg", "[]")),
        # unlikely but safeguard from invalid returns
        "Target": job.get("tgt", "unknown-target"),
        "Target-type": job.get("tgt_type", "list"),
        "User": job.get("user", "root"),
    }
    # TODO: Add Metadata support when it is merged from develop
    return ret


def _format_jid_instance(jid, job):
    """
    Format the jid correctly
    """
    ret = _format_job_instance(job)
    ret.update({"StartTime": salt.utils.jid.jid_to_time(jid)})
    return ret


def _gen_jid(cur):
    """
    Generate an unique job id
    """
    jid = salt.utils.jid.gen_jid(__opts__)
    sql = """SELECT jid FROM jids WHERE jid = %s"""
    cur.execute(sql, (jid,))
    data = cur.fetchall()
    if not data:
        return jid
    return None


def prep_jid(nocache=False, passed_jid=None):
    """
    Return a job id and prepare the job id directory
    This is the function responsible for making sure jids don't collide
    (unless its passed a jid). So do what you have to do to make sure that
    stays the case
    """
    conn = _get_conn()
    if conn is None:
        return None
    cur = conn.cursor()
    if passed_jid is None:
        jid = _gen_jid(cur)
    else:
        jid = passed_jid
    while not jid:
        log.info("jid clash, generating a new one")
        jid = _gen_jid(cur)

    cur.close()
    conn.close()
    return jid


def returner(load):
    """
    Return data to a postgres server
    """
    conn = _get_conn()
    if conn is None:
        return None
    cur = conn.cursor()
    sql = """INSERT INTO salt_returns
            (fun, jid, return, id, success)
            VALUES (%s, %s, %s, %s, %s)"""
    ret = str(load["return"])
    job_ret = {"return": ret}
    if "retcode" in load:
        job_ret["retcode"] = load["retcode"]
    if "success" in load:
        job_ret["success"] = load["success"]
    cur.execute(
        sql,
        (
            load["fun"],
            load["jid"],
            salt.utils.json.dumps(job_ret),
            load["id"],
            load.get("success"),
        ),
    )
    _close_conn(conn)


def event_return(events):
    """
    Return event to a postgres server

    Require that configuration be enabled via 'event_return'
    option in master config.
    """
    conn = _get_conn()
    if conn is None:
        return None
    cur = conn.cursor()
    for event in events:
        tag = event.get("tag", "")
        data = event.get("data", "")
        sql = """INSERT INTO salt_events
                (tag, data, master_id)
                VALUES (%s, %s, %s)"""
        cur.execute(sql, (tag, salt.utils.json.dumps(data), __opts__["id"]))
    _close_conn(conn)


def save_load(jid, clear_load, minions=None):
    """
    Save the load to the specified jid id
    """
    jid = _escape_jid(jid)
    conn = _get_conn()
    if conn is None:
        return None
    cur = conn.cursor()
    sql = (
        """INSERT INTO jids """
        """(jid, started, tgt_type, cmd, tgt, kwargs, ret, username, arg,"""
        """ fun) """
        """VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s)"""
    )

    cur.execute(
        sql,
        (
            jid,
            salt.utils.jid.jid_to_time(jid),
            str(clear_load.get("tgt_type")),
            str(clear_load.get("cmd")),
            str(clear_load.get("tgt")),
            str(clear_load.get("kwargs")),
            str(clear_load.get("ret")),
            str(clear_load.get("user")),
            str(salt.utils.json.dumps(clear_load.get("arg"))),
            str(clear_load.get("fun")),
        ),
    )
    # TODO: Add Metadata support when it is merged from develop
    _close_conn(conn)


def save_minions(jid, minions, syndic_id=None):  # pylint: disable=unused-argument
    """
    Included for API consistency
    """


def _escape_jid(jid):
    """
    Do proper formatting of the jid
    """
    jid = str(jid)
    jid = re.sub(r"'*", "", jid)
    return jid


def _build_dict(data):
    """
    Rebuild dict
    """
    result = {}
    # TODO: Add Metadata support when it is merged from develop
    result["jid"] = data[0]
    result["tgt_type"] = data[1]
    result["cmd"] = data[2]
    result["tgt"] = data[3]
    result["kwargs"] = data[4]
    result["ret"] = data[5]
    result["user"] = data[6]
    result["arg"] = data[7]
    result["fun"] = data[8]
    return result


def get_load(jid):
    """
    Return the load data that marks a specified jid
    """
    jid = _escape_jid(jid)
    conn = _get_conn()
    if conn is None:
        return None
    cur = conn.cursor()
    sql = (
        """SELECT jid, tgt_type, cmd, tgt, kwargs, ret, username, arg,"""
        """ fun FROM jids WHERE jid = %s"""
    )
    cur.execute(sql, (jid,))
    data = cur.fetchone()
    if data:
        return _build_dict(data)
    _close_conn(conn)
    return {}


def get_jid(jid):
    """
    Return the information returned when the specified job id was executed
    """
    jid = _escape_jid(jid)
    conn = _get_conn()
    if conn is None:
        return None
    cur = conn.cursor()
    sql = """SELECT id, return FROM salt_returns WHERE jid = %s"""

    cur.execute(sql, (jid,))
    data = cur.fetchall()
    ret = {}
    if data:
        for minion, full_ret in data:
            ret_data = salt.utils.json.loads(full_ret)
            if not isinstance(ret_data, dict) or "return" not in ret_data:
                # Convert the old format in which the return contains the only return data to the
                # new that is dict containing 'return' and optionally 'retcode' and 'success'.
                ret_data = {"return": ret_data}
            ret[minion] = ret_data
    _close_conn(conn)
    return ret


def get_jids():
    """
    Return a list of all job ids
    For master job cache this also formats the output and returns a string
    """
    conn = _get_conn()
    cur = conn.cursor()
    sql = (
        """SELECT """
        """jid, tgt_type, cmd, tgt, kwargs, ret, username, arg, fun """
        """FROM jids"""
    )
    if __opts__["keep_jobs"] != 0:
        sql = (
            sql
            + " WHERE started > NOW() - INTERVAL '"
            + str(__opts__["keep_jobs"])
            + "' HOUR"
        )

    cur.execute(sql)
    ret = {}
    data = cur.fetchone()
    while data:
        data_dict = _build_dict(data)
        ret[data_dict["jid"]] = _format_jid_instance(data_dict["jid"], data_dict)
        data = cur.fetchone()
    cur.close()
    conn.close()
    return ret


def clean_old_jobs():
    """
    Clean out the old jobs from the job cache
    """
    return



import salt.modules.win_psget as win_psget
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

BOOTSTRAP_PS_STR = """<?xml version="1.0" encoding="utf-8"?>
<Objects>
  <Object Type="System.Management.Automation.PSCustomObject">
    <Property Name="Name" Type="System.String">NuGet</Property>
    <Property Name="Version" Type="Microsoft.PackageManagement.Internal.Utility.Versions.FourPartVersion">
      <Property Name="Major" Type="System.UInt16">2</Property>
      <Property Name="Minor" Type="System.UInt16">8</Property>
      <Property Name="Build" Type="System.UInt16">5</Property>
      <Property Name="Revision" Type="System.UInt16">208</Property>
    </Property>
    <Property Name="ProviderPath" Type="System.String">C:\\Program Files\\PackageManagement\\ProviderAssemblies\\nuget\\2.8.5
.208\\Microsoft.PackageManagement.NuGetProvider.dll</Property>
  </Object>
</Objects>"""

AVAIL_MODULES_PS_STR = """<?xml version="1.0" encoding="utf-8"?>
<Objects>
  <Object Type="System.Management.Automation.PSCustomObject">
    <Property Name="Name" Type="System.String">ActOnCmdlets</Property>
    <Property Name="Description" Type="System.String">CData Cmdlets for Act-On</Property>
  </Object>
  <Object Type="System.Management.Automation.PSCustomObject">
    <Property Name="Name" Type="System.String">FinancialEdgeNXTCmdlets</Property>
    <Property Name="Description" Type="System.String">CData Cmdlets for Blackbaud Financial Edge NXT</Property>
  </Object>
  <Object Type="System.Management.Automation.PSCustomObject">
    <Property Name="Name" Type="System.String">GoogleCMCmdlets</Property>
    <Property Name="Description" Type="System.String">CData Cmdlets for Google Campaign Manager</Property>
  </Object>
  <Object Type="System.Management.Automation.PSCustomObject">
    <Property Name="Name" Type="System.String">DHCPMigration</Property>
    <Property Name="Description" Type="System.String">A module to copy various DHCP information from 1 server to another.</Property>
  </Object>
</Objects>"""


class WinPsgetCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.win_psget
    """

    def setup_loader_modules(self):
        return {win_psget: {}}

    def test_bootstrap(self):
        mock_read_ok = MagicMock(
            return_value={
                "pid": 78,
                "retcode": 0,
                "stderr": "",
                "stdout": BOOTSTRAP_PS_STR,
            }
        )

        with patch.dict(win_psget.__salt__, {"cmd.run_all": mock_read_ok}):
            self.assertTrue("NuGet" in win_psget.bootstrap())

    def test_avail_modules(self):
        mock_read_ok = MagicMock(
            return_value={
                "pid": 78,
                "retcode": 0,
                "stderr": "",
                "stdout": AVAIL_MODULES_PS_STR,
            }
        )

        with patch.dict(win_psget.__salt__, {"cmd.run_all": mock_read_ok}):
            self.assertTrue("DHCPMigration" in win_psget.avail_modules(False))
            self.assertTrue("DHCPMigration" in win_psget.avail_modules(True))



import pytest
import salt.states.boto_kinesis as boto_kinesis
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {boto_kinesis: {}}


def test_stream_present():
    """
    Test to ensure the kinesis stream exists.
    """
    name = "new_stream"
    retention_hours = 24
    enhanced_monitoring = ["IteratorAgeMilliseconds"]
    different_enhanced_monitoring = ["IncomingBytes"]
    num_shards = 1

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    shards = [
        {
            "ShardId": "shardId-000000000000",
            "HashKeyRange": {"EndingHashKey": "big number", "StartingHashKey": "0"},
            "SequenceNumberRange": {"StartingSequenceNumber": "bigger number"},
        }
    ]
    stream_description = {
        "HasMoreShards": False,
        "RetentionPeriodHours": retention_hours,
        "StreamName": name,
        "Shards": shards,
        "StreamARN": "",
        "EnhancedMonitoring": [{"ShardLevelMetrics": enhanced_monitoring}],
        "StreamStatus": "ACTIVE",
    }

    exists_mock = MagicMock(
        side_effect=[
            {"result": True},
            {"result": False},
            {"result": True},
            {"result": False},
        ]
    )
    get_stream_mock = MagicMock(
        return_value={"result": {"StreamDescription": stream_description}}
    )
    shard_mock = MagicMock(return_value=[0, 0, {"OpenShards": shards}])
    dict_mock = MagicMock(return_value={"result": True})
    mock_bool = MagicMock(return_value=True)
    with patch.dict(
        boto_kinesis.__salt__,
        {
            "boto_kinesis.exists": exists_mock,
            "boto_kinesis.create_stream": dict_mock,
            "boto_kinesis.get_stream_when_active": get_stream_mock,
            "boto_kinesis.get_info_for_reshard": shard_mock,
            "boto_kinesis.num_shards_matches": mock_bool,
        },
    ):
        # already present, no change required
        comt = (
            "Kinesis stream {0} already exists,\nKinesis stream {0}: retention hours"
            " did not require change, already set at {1},\nKinesis stream {0}: enhanced"
            " monitoring did not require change, already set at {2},\nKinesis stream"
            " {0}: did not require resharding, remains at {3} shards".format(
                name, retention_hours, enhanced_monitoring, num_shards
            )
        )
        ret.update({"comment": comt})
        assert (
            boto_kinesis.present(name, retention_hours, enhanced_monitoring, num_shards)
            == ret
        )

        with patch.dict(boto_kinesis.__opts__, {"test": True}):
            # not present, test environment (dry run)
            comt = "Kinesis stream {} would be created".format(name)
            ret.update({"comment": comt, "result": None})
            assert (
                boto_kinesis.present(
                    name, retention_hours, enhanced_monitoring, num_shards
                )
                == ret
            )

            # already present, changes required, test environment (dry run)
            comt = (
                "Kinesis stream {0} already exists,\n"
                "Kinesis stream {0}: retention hours would be updated to {1},\n"
                "Kinesis stream {0}: would enable enhanced monitoring for {2},\n"
                "Kinesis stream {0}: would disable enhanced monitoring for {3},\n"
                "Kinesis stream {0}: would be resharded from {4} to {5} shards".format(
                    name,
                    retention_hours + 1,
                    different_enhanced_monitoring,
                    enhanced_monitoring,
                    num_shards,
                    num_shards + 1,
                )
            )
            ret.update({"comment": comt, "result": None})
            assert (
                boto_kinesis.present(
                    name,
                    retention_hours + 1,
                    different_enhanced_monitoring,
                    num_shards + 1,
                )
                == ret
            )

        # not present, create and configure
        changes = {"new": {"name": name, "num_shards": num_shards}}

        with patch.dict(boto_kinesis.__opts__, {"test": False}):
            comt = (
                "Kinesis stream {0} successfully created,\nKinesis stream {0}:"
                " retention hours did not require change, already set at {1},\nKinesis"
                " stream {0}: enhanced monitoring did not require change, already set"
                " at {2},\nKinesis stream {0}: did not require resharding, remains at"
                " {3} shards".format(
                    name, retention_hours, enhanced_monitoring, num_shards
                )
            )
            ret.update({"comment": comt, "result": True, "changes": changes})
            assert ret == boto_kinesis.present(
                name, retention_hours, enhanced_monitoring, num_shards
            )


def test_absent():
    """
    Test to ensure the Kinesis stream does not exist.
    """
    name = "new_stream"

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    mock = MagicMock(
        side_effect=[{"result": False}, {"result": True}, {"result": True}]
    )
    mock_bool = MagicMock(return_value={"result": True})
    with patch.dict(
        boto_kinesis.__salt__,
        {"boto_kinesis.exists": mock, "boto_kinesis.delete_stream": mock_bool},
    ):
        comt = "Kinesis stream {} does not exist".format(name)
        ret.update({"comment": comt})
        assert boto_kinesis.absent(name) == ret

        with patch.dict(boto_kinesis.__opts__, {"test": True}):
            comt = "Kinesis stream {} would be deleted".format(name)
            ret.update({"comment": comt, "result": None})
            assert boto_kinesis.absent(name) == ret

        changes = {
            "new": "Stream {} deleted".format(name),
            "old": "Stream {} exists".format(name),
        }

        with patch.dict(boto_kinesis.__opts__, {"test": False}):
            comt = "Deleted stream {}".format(name)
            ret.update({"comment": comt, "result": True, "changes": changes})
            assert boto_kinesis.absent(name) == ret



"""
Return data to a Cassandra ColumnFamily

Here's an example Keyspace / ColumnFamily setup that works with this
returner::

    create keyspace salt;
    use salt;
    create column family returns
      with key_validation_class='UTF8Type'
      and comparator='UTF8Type'
      and default_validation_class='UTF8Type';

Required python modules: pycassa

  To use the cassandra returner, append '--return cassandra' to the salt command. ex:

    salt '*' test.ping --return cassandra
"""


import logging

import salt.utils.jid

try:
    import pycassa  # pylint: disable=import-error

    HAS_PYCASSA = True
except ImportError:
    HAS_PYCASSA = False

log = logging.getLogger(__name__)

__opts__ = {
    "cassandra.servers": ["localhost:9160"],
    "cassandra.keyspace": "salt",
    "cassandra.column_family": "returns",
    "cassandra.consistency_level": "ONE",
}

# Define the module's virtual name
__virtualname__ = "cassandra"


def __virtual__():
    if not HAS_PYCASSA:
        return False, "Could not import cassandra returner; pycassa is not installed."
    return __virtualname__


def returner(ret):
    """
    Return data to a Cassandra ColumnFamily
    """

    consistency_level = getattr(
        pycassa.ConsistencyLevel, __opts__["cassandra.consistency_level"]
    )

    pool = pycassa.ConnectionPool(
        __opts__["cassandra.keyspace"], __opts__["cassandra.servers"]
    )
    ccf = pycassa.ColumnFamily(
        pool,
        __opts__["cassandra.column_family"],
        write_consistency_level=consistency_level,
    )

    columns = {"fun": ret["fun"], "id": ret["id"]}
    if isinstance(ret["return"], dict):
        for key, value in ret["return"].items():
            columns["return.{}".format(key)] = str(value)
    else:
        columns["return"] = str(ret["return"])

    log.debug(columns)
    ccf.insert(ret["jid"], columns)


def prep_jid(nocache=False, passed_jid=None):  # pylint: disable=unused-argument
    """
    Do any work necessary to prepare a JID, including sending a custom id
    """
    return passed_jid if passed_jid is not None else salt.utils.jid.gen_jid(__opts__)



"""
    tests.unit.utils.scheduler.base
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""


import copy
import logging

import pytest
import salt.utils.platform
import salt.utils.schedule
from salt.modules.test import ping
from salt.utils.process import SubprocessList
from saltfactories.utils.processes import terminate_process
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture(scope="function")
def schedule(tmp_path):
    try:

        subprocess_list = SubprocessList()

        root_dir = tmp_path / "schedule-unit-tests"
        sock_dir = str(root_dir / "test-socks")

        default_config = salt.config.minion_config(None)
        default_config["conf_dir"] = str(root_dir)
        default_config["root_dir"] = str(root_dir)
        default_config["sock_dir"] = sock_dir
        default_config["pki_dir"] = str(root_dir / "pki")
        default_config["cachedir"] = str(root_dir / "cache")

        with patch("salt.utils.schedule.clean_proc_dir", MagicMock(return_value=None)):
            functions = {"test.ping": ping}
            _schedule = salt.utils.schedule.Schedule(
                copy.deepcopy(default_config),
                functions,
                returners={},
                new_instance=True,
            )

        _schedule.opts["loop_interval"] = 1
        _schedule._subprocess_list = subprocess_list

        yield _schedule

    finally:
        processes = subprocess_list.processes
        _schedule.reset()
        del _schedule
        for proc in processes:
            if proc.is_alive():
                terminate_process(proc.pid, kill_children=True, slow_stop=True)
        subprocess_list.cleanup()
        processes = subprocess_list.processes
        if processes:
            for proc in processes:
                if proc.is_alive():
                    terminate_process(proc.pid, kill_children=True, slow_stop=False)
            subprocess_list.cleanup()
        processes = subprocess_list.processes
        if processes:
            log.warning("Processes left running: %s", processes)

        del default_config
        del subprocess_list



def test_formula(salt):
    '''
    Test that the states are synced to minion
    '''
    dirs = salt('cp.list_master_dirs')
    assert 'states' in dirs

def test_wordpress_module(salt):
    '''
    Test that the wordpress dir grain was set on the minion
    '''
    wordpressdir = salt('grains.get', 'wordpressdir')
    assert salt('wordpress.is_installed', wordpressdir)



"""
tests.unit.returners.test_splunk
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for the splunk returner
"""

import json
import logging

import salt.modules.config as config
import salt.returners.splunk as splunk
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class SplunkReturnerTest(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        opts = {
            "splunk_http_forwarder": {
                "token": "TheToken",
                "indexer": "the.splunk.domain",
                "index": "TheIndex",
                "sourcetype": "TheSourceType",
            }
        }
        return {
            splunk: {"__opts__": opts, "__salt__": {"config.get": config.get}},
            config: {"__opts__": opts},
        }

    def test_verify_ssl_defaults_to_true(self):
        payload = {"some": "payload"}
        requests_post = MagicMock()
        ts = 1234565789
        host = "TheHostName"
        data = {
            "time": str(ts),
            "index": "TheIndex",
            "sourcetype": "TheSourceType",
            "event": payload,
            "host": host,
        }
        with patch(
            "salt.returners.splunk.time.time", MagicMock(return_value=ts)
        ), patch(
            "salt.returners.splunk.socket.gethostname", MagicMock(return_value=host)
        ), patch(
            "requests.post", requests_post
        ):
            splunk.returner(payload.copy())
        assert json.loads(requests_post.call_args_list[0][1]["data"]) == data
        assert requests_post.call_args_list[0][1]["verify"]
        assert requests_post.call_args_list[0][1]["headers"] == {
            "Authorization": "Splunk TheToken"
        }
        assert (
            requests_post.call_args_list[0][0][0]
            == "https://the.splunk.domain:8088/services/collector/event"
        )

    def test_verify_ssl(self):
        payload = {"some": "payload"}
        verify_ssl_values = [True, False, None]
        payload = {"some": "payload"}
        ts = 1234565789
        host = "TheHostName"
        data = {
            "time": str(ts),
            "index": "TheIndex",
            "sourcetype": "TheSourceType",
            "event": payload,
            "host": host,
        }
        for verify_ssl in verify_ssl_values:
            requests_post = MagicMock()
            with patch(
                "salt.returners.splunk.time.time", MagicMock(return_value=ts)
            ), patch(
                "salt.returners.splunk.socket.gethostname", MagicMock(return_value=host)
            ), patch(
                "requests.post", requests_post
            ), patch.dict(
                splunk.__opts__["splunk_http_forwarder"], verify_ssl=verify_ssl
            ):
                splunk.returner(payload.copy())
                assert json.loads(requests_post.call_args_list[0][1]["data"]) == data
                assert requests_post.call_args_list[0][1]["verify"] == verify_ssl
                assert requests_post.call_args_list[0][1]["headers"] == {
                    "Authorization": "Splunk TheToken"
                }
                assert (
                    requests_post.call_args_list[0][0][0]
                    == "https://the.splunk.domain:8088/services/collector/event"
                )



"""
Integration tests for the vault modules
"""
import json
import logging
import os
import subprocess
import time

import pytest
import salt.utils.path
from saltfactories.utils.processes import ProcessResult
from tests.support.runtests import RUNTIME_VARS

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.skip_if_binaries_missing("dockerd", "vault", "getent"),
]


@pytest.fixture(scope="module")
def pillar_tree(base_env_pillar_tree_root_dir, salt_minion):
    top_file = """
    base:
      '{}':
        - sdb
    """.format(
        salt_minion.id
    )
    sdb_pillar_file = """
    test_vault_pillar_sdb: sdb://sdbvault/secret/test/test_pillar_sdb/foo
    test_etcd_pillar_sdb: sdb://sdbetcd/secret/test/test_pillar_sdb/foo
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    )
    sdb_tempfile = pytest.helpers.temp_file(
        "sdb.sls", sdb_pillar_file, base_env_pillar_tree_root_dir
    )

    with top_tempfile, sdb_tempfile:
        yield


def vault_container_version_id(value):
    return "vault=={}".format(value)


@pytest.fixture(
    scope="module",
    autouse=True,
    params=["0.9.6", "1.3.1"],
    ids=vault_container_version_id,
)
def vault_container_version(request, salt_call_cli, vault_port):
    vault_version = request.param
    container_started = False
    try:
        vault_binary = salt.utils.path.which("vault")
        config = {
            "backend": {"file": {"path": "/vault/file"}},
            "default_lease_ttl": "168h",
            "max_lease_ttl": "720h",
        }
        ret = salt_call_cli.run(
            "state.single", "docker_image.present", name="vault", tag=vault_version
        )
        assert ret.exitcode == 0
        assert ret.json
        state_run = next(iter(ret.json.values()))
        assert state_run["result"] is True

        container_started = True
        attempts = 0
        env = os.environ.copy()
        env["VAULT_ADDR"] = "http://127.0.0.1:{}".format(vault_port)
        while attempts < 3:
            attempts += 1
            ret = salt_call_cli.run(
                "state.single",
                "docker_container.running",
                name="vault",
                image="vault:{}".format(vault_version),
                port_bindings="{}:8200".format(vault_port),
                environment={
                    "VAULT_DEV_ROOT_TOKEN_ID": "testsecret",
                    "VAULT_LOCAL_CONFIG": json.dumps(config),
                },
                cap_add="IPC_LOCK",
            )
            assert ret.exitcode == 0
            assert ret.json
            state_run = next(iter(ret.json.values()))
            assert state_run["result"] is True

            time.sleep(1)
            proc = subprocess.run(
                [vault_binary, "login", "token=testsecret"],
                env=env,
                check=False,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            if proc.returncode == 0:
                break
            ret = ProcessResult(
                exitcode=proc.returncode,
                stdout=proc.stdout,
                stderr=proc.stderr,
                cmdline=proc.args,
            )
            log.debug("Failed to authenticate against vault:\n%s", ret)
            time.sleep(4)
        else:
            pytest.fail("Failed to login to vault")

        proc = subprocess.run(
            [
                vault_binary,
                "policy",
                "write",
                "testpolicy",
                "{}/vault.hcl".format(RUNTIME_VARS.FILES),
            ],
            env=env,
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        if proc.returncode != 0:
            ret = ProcessResult(
                exitcode=proc.returncode,
                stdout=proc.stdout,
                stderr=proc.stderr,
                cmdline=proc.args,
            )
            log.debug("Failed to assign policy to vault:\n%s", ret)
            pytest.fail("unable to assign policy to vault")
        if vault_version == "1.3.1":
            proc = subprocess.run(
                [vault_binary, "secrets", "enable", "kv-v2"],
                env=env,
                check=False,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            ret = ProcessResult(
                exitcode=proc.returncode,
                stdout=proc.stdout,
                stderr=proc.stderr,
                cmdline=proc.args,
            )
            if proc.returncode != 0:
                log.debug("Failed to enable kv-v2:\n%s", ret)
                pytest.fail("Could not enable kv-v2")

            if "path is already in use at kv-v2/" in proc.stdout:
                pass
            elif "Success" in proc.stdout:
                pass
            else:
                log.debug("Failed to enable kv-v2:\n%s", ret)
                pytest.fail("Could not enable kv-v2 {}".format(proc.stdout))
        yield vault_version
    finally:
        if container_started:
            ret = salt_call_cli.run(
                "state.single", "docker_container.stopped", name="vault"
            )
            assert ret.exitcode == 0
            assert ret.json
            state_run = next(iter(ret.json.values()))
            assert state_run["result"] is True
            ret = salt_call_cli.run(
                "state.single", "docker_container.absent", name="vault"
            )
            assert ret.exitcode == 0
            assert ret.json
            state_run = next(iter(ret.json.values()))
            assert state_run["result"] is True


@pytest.mark.slow_test
def test_sdb(salt_call_cli):
    ret = salt_call_cli.run(
        "sdb.set", uri="sdb://sdbvault/secret/test/test_sdb/foo", value="bar"
    )
    assert ret.exitcode == 0
    assert ret.json is True
    ret = salt_call_cli.run("sdb.get", uri="sdb://sdbvault/secret/test/test_sdb/foo")
    assert ret.exitcode == 0
    assert ret.json
    assert ret.json == "bar"


@pytest.mark.slow_test
def test_sdb_runner(salt_run_cli):
    ret = salt_run_cli.run(
        "sdb.set", uri="sdb://sdbvault/secret/test/test_sdb_runner/foo", value="bar"
    )
    assert ret.exitcode == 0
    assert ret.json is True
    ret = salt_run_cli.run(
        "sdb.get", uri="sdb://sdbvault/secret/test/test_sdb_runner/foo"
    )
    assert ret.exitcode == 0
    assert ret.stdout
    assert ret.stdout == "bar"


@pytest.mark.slow_test
def test_config(salt_call_cli, pillar_tree):
    ret = salt_call_cli.run(
        "sdb.set", uri="sdb://sdbvault/secret/test/test_pillar_sdb/foo", value="bar"
    )
    assert ret.exitcode == 0
    assert ret.json is True
    ret = salt_call_cli.run("config.get", "test_vault_pillar_sdb")
    assert ret.exitcode == 0
    assert ret.json
    assert ret.json == "bar"


@pytest.mark.slow_test
def test_sdb_kv2_kvv2_path_local(salt_call_cli, vault_container_version):
    if vault_container_version != "1.3.1":
        pytest.skip("Test not applicable to vault {}".format(vault_container_version))
    ret = salt_call_cli.run(
        "sdb.set", uri="sdb://sdbvault/kv-v2/test/test_sdb/foo", value="bar"
    )
    assert ret.exitcode == 0
    assert ret.json is True
    ret = salt_call_cli.run(
        "--local", "sdb.get", "sdb://sdbvault/kv-v2/test/test_sdb/foo"
    )
    assert ret.json
    assert ret.json == "bar"



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import os

import salt.modules.kmod as kmod
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf


class KmodTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.kmod
    """

    def setup_loader_modules(self):
        return {kmod: {}}

    # 'available' function tests: 1

    def test_available(self):
        """
        Tests return a list of all available kernel modules
        """
        with patch("salt.modules.kmod.available", MagicMock(return_value=["kvm"])):
            self.assertEqual(["kvm"], kmod.available())

    # 'check_available' function tests: 1

    def test_check_available(self):
        """
        Tests if the specified kernel module is available
        """
        with patch("salt.modules.kmod.available", MagicMock(return_value=["kvm"])):
            self.assertTrue(kmod.check_available("kvm"))

    # 'lsmod' function tests: 1

    def test_lsmod(self):
        """
        Tests return information about currently loaded modules
        """
        mock_ret = [{"size": 100, "module": None, "depcount": 10, "deps": None}]
        with patch("salt.modules.kmod.lsmod", MagicMock(return_value=mock_ret)):
            mock_cmd = MagicMock(return_value=1)
            with patch.dict(kmod.__salt__, {"cmd.run": mock_cmd}):
                self.assertListEqual(mock_ret, kmod.lsmod())

    # 'mod_list' function tests: 1

    @skipIf(not os.path.isfile("/etc/modules"), "/etc/modules not present")
    def test_mod_list(self):
        """
        Tests return a list of the loaded module names
        """
        with patch(
            "salt.modules.kmod._get_modules_conf",
            MagicMock(return_value="/etc/modules"),
        ):
            with patch(
                "salt.modules.kmod._strip_module_name", MagicMock(return_value="lp")
            ):
                self.assertListEqual(["lp"], kmod.mod_list(True))

        mock_ret = [{"size": 100, "module": None, "depcount": 10, "deps": None}]
        with patch("salt.modules.kmod.lsmod", MagicMock(return_value=mock_ret)):
            self.assertListEqual([None], kmod.mod_list(False))

    # 'load' function tests: 1

    def test_load(self):
        """
        Tests to loads specified kernel module.
        """
        mod = "cheese"
        err_msg = "Module too moldy, refusing to load"
        mock_persist = MagicMock(return_value={mod})
        mock_lsmod = MagicMock(
            return_value=[{"size": 100, "module": None, "depcount": 10, "deps": None}]
        )
        mock_run_all_0 = MagicMock(return_value={"retcode": 0})
        mock_run_all_1 = MagicMock(return_value={"retcode": 1, "stderr": err_msg})

        with patch("salt.modules.kmod._set_persistent_module", mock_persist):
            with patch("salt.modules.kmod.lsmod", mock_lsmod):
                with patch.dict(kmod.__salt__, {"cmd.run_all": mock_run_all_0}):
                    self.assertEqual([mod], kmod.load(mod, True))

                with patch.dict(kmod.__salt__, {"cmd.run_all": mock_run_all_1}):
                    self.assertEqual(
                        "Error loading module {}: {}".format(mod, err_msg),
                        kmod.load(mod),
                    )

    # 'is_loaded' function tests: 1

    def test_is_loaded(self):
        """
        Tests if specified kernel module is loaded.
        """
        with patch("salt.modules.kmod.mod_list", MagicMock(return_value={"lp"})):
            self.assertTrue(kmod.is_loaded("lp"))

    # 'remove' function tests: 1

    def test_remove(self):
        """
        Tests to remove the specified kernel module
        """
        mod = "cheese"
        err_msg = "Cannot find module: it has been eaten"
        mock_persist = MagicMock(return_value={mod})
        mock_lsmod = MagicMock(
            return_value=[{"size": 100, "module": None, "depcount": 10, "deps": None}]
        )
        mock_run_all_0 = MagicMock(return_value={"retcode": 0})
        mock_run_all_1 = MagicMock(return_value={"retcode": 1, "stderr": err_msg})

        with patch("salt.modules.kmod._remove_persistent_module", mock_persist):
            with patch("salt.modules.kmod.lsmod", mock_lsmod):
                with patch.dict(kmod.__salt__, {"cmd.run_all": mock_run_all_0}):
                    self.assertEqual([mod], kmod.remove(mod, True))

                    self.assertEqual([], kmod.remove(mod))

                with patch.dict(kmod.__salt__, {"cmd.run_all": mock_run_all_1}):
                    self.assertEqual(
                        "Error removing module {}: {}".format(mod, err_msg),
                        kmod.remove(mod, True),
                    )



"""
Tests for the Reg State
"""

import logging

import pytest
import salt.utils.platform
import salt.utils.win_reg as reg
from tests.support.case import ModuleCase
from tests.support.helpers import random_string
from tests.support.mixins import SaltReturnAssertsMixin
from tests.support.unit import skipIf

log = logging.getLogger(__name__)

UNICODE_VALUE_NAME = "Unicode Key \N{TRADE MARK SIGN}"
UNICODE_VALUE = (
    "Unicode Value \N{COPYRIGHT SIGN},\N{TRADE MARK SIGN},\N{REGISTERED SIGN}"
)
FAKE_KEY = "SOFTWARE\\{}".format(random_string("SaltTesting-", lowercase=False))


@pytest.mark.destructive_test
@skipIf(not salt.utils.platform.is_windows(), "Windows Specific Test")
@pytest.mark.windows_whitelisted
class RegTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Reg state module tests
    These tests are destructive as the modify the registry
    """

    def tearDown(self):
        reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY)
        reg.delete_key_recursive(hive="HKLM", key=FAKE_KEY, use_32bit_registry=True)

    @pytest.mark.slow_test
    def test_present_reg_sz(self):
        """
        Testing reg.present with REG_SZ
        """
        log.debug("Testing reg.present with REG_SZ")
        # default type is 'REG_SZ'
        # Does the state return the correct data
        ret = self.run_state(
            "reg.present",
            name="HKLM\\{}".format(FAKE_KEY),
            vname="test_reg_sz",
            vdata="fake string data",
        )
        expected = {
            "reg": {
                "Added": {
                    "Entry": "test_reg_sz",
                    "Inheritance": True,
                    "Key": "HKLM\\{}".format(FAKE_KEY),
                    "Owner": None,
                    "Perms": {"Deny": None, "Grant": None},
                    "Value": "fake string data",
                }
            }
        }
        self.assertSaltStateChangesEqual(ret, expected)

        # Is the value actually set
        ret = reg.read_value(hive="HKLM", key=FAKE_KEY, vname="test_reg_sz")
        expected = {
            "vtype": "REG_SZ",
            "vname": "test_reg_sz",
            "success": True,
            "hive": "HKLM",
            "vdata": "fake string data",
            "key": FAKE_KEY,
        }
        self.assertEqual(ret, expected)

    @pytest.mark.slow_test
    def test_present_reg_sz_unicode_value(self):
        """
        Testing reg.present with REG_SZ and a unicode value
        """
        log.debug("Testing reg.present with REG_SZ and a unicode value")
        # default type is 'REG_SZ'
        # Does the state return the correct data
        ret = self.run_state(
            "reg.present",
            name="HKLM\\{}".format(FAKE_KEY),
            vname="test_reg_sz",
            vdata=UNICODE_VALUE,
        )
        expected = {
            "reg": {
                "Added": {
                    "Entry": "test_reg_sz",
                    "Inheritance": True,
                    "Key": "HKLM\\{}".format(FAKE_KEY),
                    "Owner": None,
                    "Perms": {"Deny": None, "Grant": None},
                    "Value": UNICODE_VALUE,
                }
            }
        }
        self.assertSaltStateChangesEqual(ret, expected)

        # Is the value actually set
        ret = reg.read_value(hive="HKLM", key=FAKE_KEY, vname="test_reg_sz")
        expected = {
            "vtype": "REG_SZ",
            "vname": "test_reg_sz",
            "success": True,
            "hive": "HKLM",
            "vdata": UNICODE_VALUE,
            "key": FAKE_KEY,
        }
        self.assertEqual(ret, expected)

    @pytest.mark.slow_test
    def test_present_reg_sz_unicode_default_value(self):
        """
        Testing reg.present with REG_SZ and a unicode default value
        """
        log.debug("Testing reg.present with REG_SZ and a unicode default value")
        # default type is 'REG_SZ'
        # Does the state return the correct data
        ret = self.run_state(
            "reg.present", name="HKLM\\{}".format(FAKE_KEY), vdata=UNICODE_VALUE
        )
        expected = {
            "reg": {
                "Added": {
                    "Entry": "(Default)",
                    "Inheritance": True,
                    "Key": "HKLM\\{}".format(FAKE_KEY),
                    "Owner": None,
                    "Perms": {"Deny": None, "Grant": None},
                    "Value": UNICODE_VALUE,
                }
            }
        }
        self.assertSaltStateChangesEqual(ret, expected)

        # Is the value actually set
        ret = reg.read_value(hive="HKLM", key=FAKE_KEY)

        expected = {
            "vtype": "REG_SZ",
            "vname": "(Default)",
            "success": True,
            "hive": "HKLM",
            "vdata": UNICODE_VALUE,
            "key": FAKE_KEY,
        }
        self.assertEqual(ret, expected)

    @pytest.mark.slow_test
    def test_present_reg_sz_unicode_value_name(self):
        """
        Testing reg.present with REG_SZ and a unicode value name
        """
        log.debug("Testing reg.present with REG_SZ and a unicode value name")
        # default type is 'REG_SZ'
        # Does the state return the correct data
        ret = self.run_state(
            "reg.present",
            name="HKLM\\{}".format(FAKE_KEY),
            vname=UNICODE_VALUE_NAME,
            vdata="fake string data",
        )
        expected = {
            "reg": {
                "Added": {
                    "Entry": UNICODE_VALUE_NAME,
                    "Inheritance": True,
                    "Key": "HKLM\\{}".format(FAKE_KEY),
                    "Owner": None,
                    "Perms": {"Deny": None, "Grant": None},
                    "Value": "fake string data",
                }
            }
        }
        self.assertSaltStateChangesEqual(ret, expected)

        # Is the value actually set
        ret = reg.read_value(hive="HKLM", key=FAKE_KEY, vname=UNICODE_VALUE_NAME)

        expected = {
            "vtype": "REG_SZ",
            "vname": UNICODE_VALUE_NAME,
            "success": True,
            "hive": "HKLM",
            "vdata": "fake string data",
            "key": FAKE_KEY,
        }
        self.assertEqual(ret, expected)

    @pytest.mark.slow_test
    def test_present_reg_binary(self):
        """
        Testing reg.present with REG_BINARY
        """
        test_data = "Salty Test"
        log.debug("Testing reg.present with REG_BINARY")
        # default type is 'REG_SZ'
        # Does the state return the correct data
        ret = self.run_state(
            "reg.present",
            name="HKLM\\{}".format(FAKE_KEY),
            vname="test_reg_binary",
            vtype="REG_BINARY",
            vdata=test_data,
        )
        expected = {
            "reg": {
                "Added": {
                    "Entry": "test_reg_binary",
                    "Inheritance": True,
                    "Key": "HKLM\\{}".format(FAKE_KEY),
                    "Owner": None,
                    "Perms": {"Deny": None, "Grant": None},
                    "Value": test_data,
                }
            }
        }
        self.assertSaltStateChangesEqual(ret, expected)

        # Is the value actually set
        ret = reg.read_value(hive="HKLM", key=FAKE_KEY, vname="test_reg_binary")
        expected = {
            "vtype": "REG_BINARY",
            "vname": "test_reg_binary",
            "success": True,
            "hive": "HKLM",
            "vdata": test_data.encode("utf-8"),
            "key": FAKE_KEY,
        }
        self.assertEqual(ret, expected)

    @pytest.mark.slow_test
    def test_present_reg_multi_sz(self):
        """
        Testing reg.present with REG_MULTI_SZ
        """
        log.debug("Testing reg.present with REG_MULTI_SZ")
        # default type is 'REG_SZ'
        # Does the state return the correct data
        ret = self.run_state(
            "reg.present",
            name="HKLM\\{}".format(FAKE_KEY),
            vname="test_reg_multi_sz",
            vtype="REG_MULTI_SZ",
            vdata=["item1", "item2"],
        )
        expected = {
            "reg": {
                "Added": {
                    "Entry": "test_reg_multi_sz",
                    "Inheritance": True,
                    "Key": "HKLM\\{}".format(FAKE_KEY),
                    "Owner": None,
                    "Perms": {"Deny": None, "Grant": None},
                    "Value": ["item1", "item2"],
                }
            }
        }
        self.assertSaltStateChangesEqual(ret, expected)

        # Is the value actually set
        ret = reg.read_value(hive="HKLM", key=FAKE_KEY, vname="test_reg_multi_sz")
        expected = {
            "vtype": "REG_MULTI_SZ",
            "vname": "test_reg_multi_sz",
            "success": True,
            "hive": "HKLM",
            "vdata": ["item1", "item2"],
            "key": FAKE_KEY,
        }
        self.assertEqual(ret, expected)

    @pytest.mark.slow_test
    def test_present_32_bit(self):
        """
        Testing reg.present with REG_SZ using 32bit registry
        """
        log.debug("Testing reg.present with REG_SZ using 32bit registry")
        # default type is 'REG_SZ'
        # Does the state return the correct data
        ret = self.run_state(
            "reg.present",
            name="HKLM\\{}".format(FAKE_KEY),
            vname="test_reg_sz",
            vdata="fake string data",
            use_32bit_registry=True,
        )

        expected = {
            "reg": {
                "Added": {
                    "Entry": "test_reg_sz",
                    "Inheritance": True,
                    "Key": "HKLM\\{}".format(FAKE_KEY),
                    "Owner": None,
                    "Perms": {"Deny": None, "Grant": None},
                    "Value": "fake string data",
                }
            }
        }
        self.assertSaltStateChangesEqual(ret, expected)

        # Is the value actually set
        ret = reg.read_value(
            hive="HKLM", key=FAKE_KEY, vname="test_reg_sz", use_32bit_registry=True
        )
        expected = {
            "vtype": "REG_SZ",
            "vname": "test_reg_sz",
            "success": True,
            "hive": "HKLM",
            "vdata": "fake string data",
            "key": FAKE_KEY,
        }
        self.assertEqual(ret, expected)



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import os

import salt.states.hg as hg
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class HgTestCase(TestCase, LoaderModuleMockMixin):
    """
    Validate the svn state
    """

    def setup_loader_modules(self):
        return {hg: {}}

    def test_latest(self):
        """
        Test to Make sure the repository is cloned to
        the given directory and is up to date
        """
        ret = {"changes": {}, "comment": "", "name": "salt", "result": True}
        mock = MagicMock(return_value=True)
        with patch.object(hg, "_fail", mock):
            self.assertTrue(hg.latest("salt"))

        mock = MagicMock(side_effect=[False, True, False, False, False, False])
        with patch.object(os.path, "isdir", mock):
            mock = MagicMock(return_value=True)
            with patch.object(hg, "_handle_existing", mock):
                self.assertTrue(hg.latest("salt", target="c:\\salt"))

            with patch.dict(hg.__opts__, {"test": True}):
                mock = MagicMock(return_value=True)
                with patch.object(hg, "_neutral_test", mock):
                    self.assertTrue(hg.latest("salt", target="c:\\salt"))

            with patch.dict(hg.__opts__, {"test": False}):
                mock = MagicMock(return_value=True)
                with patch.object(hg, "_clone_repo", mock):
                    self.assertDictEqual(hg.latest("salt", target="c:\\salt"), ret)

    def test_latest_update_changes(self):
        """
        Test to make sure we don't update even if we have changes
        """
        ret = {"changes": {}, "comment": "", "name": "salt", "result": True}
        revision_mock = MagicMock(return_value="abcdef")
        pull_mock = MagicMock(return_value="Blah.")
        update_mock = MagicMock()

        with patch.dict(
            hg.__salt__,
            {
                "hg.revision": revision_mock,
                "hg.pull": pull_mock,
                "hg.update": update_mock,
            },
        ):
            mock = MagicMock(side_effect=[True, True])
            with patch.object(os.path, "isdir", mock):
                mock = MagicMock(return_value=True)
                with patch.dict(hg.__opts__, {"test": False}):
                    with patch.object(hg, "_clone_repo", mock):
                        self.assertDictEqual(
                            hg.latest("salt", target="c:\\salt", update_head=True), ret
                        )
                        assert update_mock.called

    def test_latest_no_update_changes(self):
        """
        Test to make sure we don't update even if we have changes
        """
        ret = {
            "changes": {},
            "comment": (
                "Update is probably required but update_head=False so we will skip"
                " updating."
            ),
            "name": "salt",
            "result": True,
        }
        revision_mock = MagicMock(return_value="abcdef")
        pull_mock = MagicMock(return_value="Blah.")
        update_mock = MagicMock()

        with patch.dict(
            hg.__salt__,
            {
                "hg.revision": revision_mock,
                "hg.pull": pull_mock,
                "hg.update": update_mock,
            },
        ):
            mock = MagicMock(side_effect=[True, True])
            with patch.object(os.path, "isdir", mock):
                mock = MagicMock(return_value=True)
                with patch.dict(hg.__opts__, {"test": False}):
                    with patch.object(hg, "_clone_repo", mock):
                        self.assertDictEqual(
                            hg.latest("salt", target="c:\\salt", update_head=False), ret
                        )
                        assert not update_mock.called

    def test_latest_no_update_no_changes(self):
        """
        Test to Make sure the repository is cloned to
        the given directory and is up to date
        """
        ret = {
            "changes": {},
            "comment": "No changes found and update_head=False so will skip updating.",
            "name": "salt",
            "result": True,
        }
        revision_mock = MagicMock(return_value="abcdef")
        pull_mock = MagicMock(return_value="Blah no changes found.")
        update_mock = MagicMock()

        with patch.dict(
            hg.__salt__,
            {
                "hg.revision": revision_mock,
                "hg.pull": pull_mock,
                "hg.update": update_mock,
            },
        ):
            mock = MagicMock(side_effect=[True, True])
            with patch.object(os.path, "isdir", mock):
                mock = MagicMock(return_value=True)
                with patch.dict(hg.__opts__, {"test": False}):
                    with patch.object(hg, "_clone_repo", mock):
                        self.assertDictEqual(
                            hg.latest("salt", target="c:\\salt", update_head=False), ret
                        )
                        assert not update_mock.called



"""
These only test the provider selection and verification logic, they do not init
any remotes.
"""


import os
import shutil
from time import time

import salt.fileserver.gitfs
import salt.utils.files
import salt.utils.gitfs
import salt.utils.platform
import tests.support.paths
from salt.exceptions import FileserverConfigError
from tests.support.mixins import AdaptedConfigurationTestCaseMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

try:
    HAS_PYGIT2 = (
        salt.utils.gitfs.PYGIT2_VERSION >= salt.utils.gitfs.PYGIT2_MINVER
        and salt.utils.gitfs.LIBGIT2_VERSION >= salt.utils.gitfs.LIBGIT2_MINVER
    )
except AttributeError:
    HAS_PYGIT2 = False


if HAS_PYGIT2:
    import pygit2


def _clear_instance_map():
    try:
        del salt.utils.gitfs.GitFS.instance_map[
            salt.ext.tornado.ioloop.IOLoop.current()
        ]
    except KeyError:
        pass


class TestGitBase(TestCase, AdaptedConfigurationTestCaseMixin):
    def setUp(self):
        class MockedProvider(
            salt.utils.gitfs.GitProvider
        ):  # pylint: disable=abstract-method
            def __init__(
                self,
                opts,
                remote,
                per_remote_defaults,
                per_remote_only,
                override_params,
                cache_root,
                role="gitfs",
            ):
                self.provider = "mocked"
                self.fetched = False
                super().__init__(
                    opts,
                    remote,
                    per_remote_defaults,
                    per_remote_only,
                    override_params,
                    cache_root,
                    role,
                )

            def init_remote(self):
                self.repo = True
                new = False
                return new

            def envs(self):
                return ["base"]

            def fetch(self):
                self.fetched = True

        git_providers = {
            "mocked": MockedProvider,
        }
        gitfs_remotes = ["file://repo1.git", {"file://repo2.git": [{"name": "repo2"}]}]
        self.opts = self.get_temp_config(
            "master", gitfs_remotes=gitfs_remotes, verified_gitfs_provider="mocked"
        )
        self.main_class = salt.utils.gitfs.GitFS(
            self.opts,
            self.opts["gitfs_remotes"],
            per_remote_overrides=salt.fileserver.gitfs.PER_REMOTE_OVERRIDES,
            per_remote_only=salt.fileserver.gitfs.PER_REMOTE_ONLY,
            git_providers=git_providers,
        )

    @classmethod
    def setUpClass(cls):
        # Clear the instance map so that we make sure to create a new instance
        # for this test class.
        _clear_instance_map()

    def tearDown(self):
        # Providers are preserved with GitFS's instance_map
        for remote in self.main_class.remotes:
            remote.fetched = False
        del self.main_class

    def test_update_all(self):
        self.main_class.update()
        self.assertEqual(len(self.main_class.remotes), 2, "Wrong number of remotes")
        self.assertTrue(self.main_class.remotes[0].fetched)
        self.assertTrue(self.main_class.remotes[1].fetched)

    def test_update_by_name(self):
        self.main_class.update("repo2")
        self.assertEqual(len(self.main_class.remotes), 2, "Wrong number of remotes")
        self.assertFalse(self.main_class.remotes[0].fetched)
        self.assertTrue(self.main_class.remotes[1].fetched)

    def test_update_by_id_and_name(self):
        self.main_class.update([("file://repo1.git", None)])
        self.assertEqual(len(self.main_class.remotes), 2, "Wrong number of remotes")
        self.assertTrue(self.main_class.remotes[0].fetched)
        self.assertFalse(self.main_class.remotes[1].fetched)


class TestGitFSProvider(TestCase):
    def setUp(self):
        self.opts = {"cachedir": "/tmp/gitfs-test-cache"}

    def tearDown(self):
        self.opts = None

    def test_provider_case_insensitive(self):
        """
        Ensure that both lowercase and non-lowercase values are supported
        """
        provider = "GitPython"
        for role_name, role_class in (
            ("gitfs", salt.utils.gitfs.GitFS),
            ("git_pillar", salt.utils.gitfs.GitPillar),
            ("winrepo", salt.utils.gitfs.WinRepo),
        ):

            key = "{}_provider".format(role_name)
            with patch.object(
                role_class, "verify_gitpython", MagicMock(return_value=True)
            ):
                with patch.object(
                    role_class, "verify_pygit2", MagicMock(return_value=False)
                ):
                    args = [self.opts, {}]
                    kwargs = {"init_remotes": False}
                    if role_name == "winrepo":
                        kwargs["cache_root"] = "/tmp/winrepo-dir"
                    with patch.dict(self.opts, {key: provider}):
                        # Try to create an instance with uppercase letters in
                        # provider name. If it fails then a
                        # FileserverConfigError will be raised, so no assert is
                        # necessary.
                        role_class(*args, **kwargs)
                    # Now try to instantiate an instance with all lowercase
                    # letters. Again, no need for an assert here.
                    role_class(*args, **kwargs)

    def test_valid_provider(self):
        """
        Ensure that an invalid provider is not accepted, raising a
        FileserverConfigError.
        """

        def _get_mock(verify, provider):
            """
            Return a MagicMock with the desired return value
            """
            return MagicMock(return_value=verify.endswith(provider))

        for role_name, role_class in (
            ("gitfs", salt.utils.gitfs.GitFS),
            ("git_pillar", salt.utils.gitfs.GitPillar),
            ("winrepo", salt.utils.gitfs.WinRepo),
        ):
            key = "{}_provider".format(role_name)
            for provider in salt.utils.gitfs.GIT_PROVIDERS:
                verify = "verify_gitpython"
                mock1 = _get_mock(verify, provider)
                with patch.object(role_class, verify, mock1):
                    verify = "verify_pygit2"
                    mock2 = _get_mock(verify, provider)
                    with patch.object(role_class, verify, mock2):
                        args = [self.opts, {}]
                        kwargs = {"init_remotes": False}
                        if role_name == "winrepo":
                            kwargs["cache_root"] = "/tmp/winrepo-dir"

                        with patch.dict(self.opts, {key: provider}):
                            role_class(*args, **kwargs)

                        with patch.dict(self.opts, {key: "foo"}):
                            # Set the provider name to a known invalid provider
                            # and make sure it raises an exception.
                            self.assertRaises(
                                FileserverConfigError, role_class, *args, **kwargs
                            )


@skipIf(not HAS_PYGIT2, "This host lacks proper pygit2 support")
@skipIf(
    salt.utils.platform.is_windows(),
    "Skip Pygit2 on windows, due to pygit2 access error on windows",
)
class TestPygit2(TestCase):
    def _prepare_remote_repository(self, path):
        shutil.rmtree(path, ignore_errors=True)

        filecontent = "This is an empty README file"
        filename = "README"

        signature = pygit2.Signature(
            "Dummy Commiter", "dummy@dummy.com", int(time()), 0
        )

        repository = pygit2.init_repository(path, False)
        builder = repository.TreeBuilder()
        tree = builder.write()
        commit = repository.create_commit(
            "HEAD", signature, signature, "Create master branch", tree, []
        )
        repository.create_reference("refs/tags/simple_tag", commit)

        with salt.utils.files.fopen(
            os.path.join(repository.workdir, filename), "w"
        ) as file:
            file.write(filecontent)

        blob = repository.create_blob_fromworkdir(filename)
        builder = repository.TreeBuilder()
        builder.insert(filename, blob, pygit2.GIT_FILEMODE_BLOB)
        tree = builder.write()

        repository.index.read()
        repository.index.add(filename)
        repository.index.write()

        commit = repository.create_commit(
            "HEAD",
            signature,
            signature,
            "Added a README",
            tree,
            [repository.head.target],
        )
        repository.create_tag(
            "annotated_tag", commit, pygit2.GIT_OBJ_COMMIT, signature, "some message"
        )

    def _prepare_cache_repository(self, remote, cache):
        opts = {
            "cachedir": cache,
            "__role": "minion",
            "gitfs_disable_saltenv_mapping": False,
            "gitfs_base": "master",
            "gitfs_insecure_auth": False,
            "gitfs_mountpoint": "",
            "gitfs_passphrase": "",
            "gitfs_password": "",
            "gitfs_privkey": "",
            "gitfs_provider": "pygit2",
            "gitfs_pubkey": "",
            "gitfs_ref_types": ["branch", "tag", "sha"],
            "gitfs_refspecs": [
                "+refs/heads/*:refs/remotes/origin/*",
                "+refs/tags/*:refs/tags/*",
            ],
            "gitfs_root": "",
            "gitfs_saltenv_blacklist": [],
            "gitfs_saltenv_whitelist": [],
            "gitfs_ssl_verify": True,
            "gitfs_update_interval": 3,
            "gitfs_user": "",
            "verified_gitfs_provider": "pygit2",
        }
        per_remote_defaults = {
            "base": "master",
            "disable_saltenv_mapping": False,
            "insecure_auth": False,
            "ref_types": ["branch", "tag", "sha"],
            "passphrase": "",
            "mountpoint": "",
            "password": "",
            "privkey": "",
            "pubkey": "",
            "refspecs": [
                "+refs/heads/*:refs/remotes/origin/*",
                "+refs/tags/*:refs/tags/*",
            ],
            "root": "",
            "saltenv_blacklist": [],
            "saltenv_whitelist": [],
            "ssl_verify": True,
            "update_interval": 60,
            "user": "",
        }
        per_remote_only = ("all_saltenvs", "name", "saltenv")
        override_params = tuple(per_remote_defaults.keys())
        cache_root = os.path.join(cache, "gitfs")
        role = "gitfs"
        shutil.rmtree(cache_root, ignore_errors=True)
        provider = salt.utils.gitfs.Pygit2(
            opts,
            remote,
            per_remote_defaults,
            per_remote_only,
            override_params,
            cache_root,
            role,
        )
        return provider

    def test_checkout(self):
        remote = os.path.join(tests.support.paths.TMP, "pygit2-repo")
        cache = os.path.join(tests.support.paths.TMP, "pygit2-repo-cache")
        self._prepare_remote_repository(remote)
        provider = self._prepare_cache_repository(remote, cache)
        provider.remotecallbacks = None
        provider.credentials = None
        provider.init_remote()
        provider.fetch()
        provider.branch = "master"
        self.assertIn(provider.cachedir, provider.checkout())
        provider.branch = "simple_tag"
        self.assertIn(provider.cachedir, provider.checkout())
        provider.branch = "annotated_tag"
        self.assertIn(provider.cachedir, provider.checkout())
        provider.branch = "does_not_exist"
        self.assertIsNone(provider.checkout())



"""
    :codeauthor: Gareth J. Greenaway <gareth@saltstack.com>
"""

import salt.modules.napalm_network as net_mod
import salt.states.netconfig as netconfig
import salt.utils.files
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class NetconfigTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.netconfig
    """

    def setup_loader_modules(self):
        state_loader_globals = {
            "__env__": "base",
            "__salt__": {"net.replace_pattern": net_mod.replace_pattern},
        }
        module_loader_globals = {
            "__env__": "base",
            "__salt__": {
                "net.replace_pattern": net_mod.replace_pattern,
                "net.load_config": net_mod.load_config,
            },
        }
        return {netconfig: state_loader_globals, net_mod: module_loader_globals}

    def test_replace_pattern_test_is_true(self):
        """
        Test to replace_pattern to ensure that test=True
        is being passed correctly.
        """
        name = "name"
        pattern = "OLD-POLICY-NAME"
        repl = "new-policy-name"

        mock = MagicMock()
        mock_net_replace_pattern = MagicMock()
        mock_loaded_ret = MagicMock()

        with patch.dict(netconfig.__salt__, {"config.merge": mock}):
            with patch.dict(
                netconfig.__salt__, {"net.replace_pattern": mock_net_replace_pattern}
            ):
                with patch.object(salt.utils.napalm, "loaded_ret", mock_loaded_ret):
                    # Test if test=True is passed as argument to state.apply
                    with patch.dict(netconfig.__opts__, {"test": True}):
                        netconfig.replace_pattern(name, pattern, repl)

                        # Get the args and kwargs from the mocked call net.replace_pattern
                        args, kwargs = mock_net_replace_pattern.call_args_list[0]

                        # Verify that the keyword argument is True
                        self.assertTrue(kwargs["test"])

                        # Get the args and kwargs from the mocked call to salt.utils.napalm.loaded_ret
                        args, kwargs = mock_loaded_ret.call_args_list[0]

                        # Verify that the third positional argument is True
                        self.assertTrue(args[2])

                    # Test if test=True is passed as argument to state directly
                    netconfig.replace_pattern(name, pattern, repl, test=True)

                    # Get the args and kwargs from the mocked call net.replace_pattern
                    args, kwargs = mock_net_replace_pattern.call_args_list[0]

                    # Verify that the keyword argument is True
                    self.assertTrue(kwargs["test"])

                    # Get the args and kwargs from the mocked call to salt.utils.napalm.loaded_ret
                    args, kwargs = mock_loaded_ret.call_args_list[0]

                    # Verify that the third positional argument is True
                    self.assertTrue(args[2])

    def test_managed_test_is_true(self):
        """
        Test to managed to ensure that test=True
        is being passed correctly.
        """
        name = "name"

        mock = MagicMock()
        mock_update_config = MagicMock()

        with patch.dict(netconfig.__salt__, {"config.merge": mock}):
            with patch.object(netconfig, "_update_config", mock_update_config):
                # Test if test=True is passed as argument to state.apply
                with patch.dict(netconfig.__opts__, {"test": True}):
                    netconfig.managed(name)

                    # Get the args and kwargs from the mocked call net.replace_pattern
                    args, kwargs = mock_update_config.call_args_list[0]

                    # Verify that the keyword argument is True
                    self.assertTrue(kwargs["test"])

                # Test if test=True is passed as argument to state directly
                netconfig.managed(name, test=True)

                # Get the args and kwargs from the mocked call net.replace_pattern
                args, kwargs = mock_update_config.call_args_list[0]

                # Verify that the keyword argument is True
                self.assertTrue(kwargs["test"])



import os
import shutil
import stat

import pytest
import salt.utils.files
import salt.utils.stringutils
from tests.support.case import ShellCase
from tests.support.runtests import RUNTIME_VARS


class AutosignGrainsTest(ShellCase):
    """
    Test autosigning minions based on grain values.
    """

    def setUp(self):
        # all read, only owner write
        self.autosign_file_permissions = (
            stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IWUSR
        )
        self.autosign_file_path = os.path.join(RUNTIME_VARS.TMP, "autosign_file")
        shutil.copyfile(
            os.path.join(RUNTIME_VARS.FILES, "autosign_grains", "autosign_file"),
            self.autosign_file_path,
        )
        os.chmod(self.autosign_file_path, self.autosign_file_permissions)

        self.run_key("-d minion -y")
        self.run_call(
            "test.ping -l quiet"
        )  # get minion to try to authenticate itself again

        if "minion" in self.run_key("-l acc"):
            self.tearDown()
            self.skipTest("Could not deauthorize minion")
        if "minion" not in self.run_key("-l un"):
            self.tearDown()
            self.skipTest("minion did not try to reauthenticate itself")

        self.autosign_grains_dir = os.path.join(self.master_opts["autosign_grains_dir"])
        if not os.path.isdir(self.autosign_grains_dir):
            os.makedirs(self.autosign_grains_dir)

    def tearDown(self):
        shutil.copyfile(
            os.path.join(RUNTIME_VARS.FILES, "autosign_file"), self.autosign_file_path
        )
        os.chmod(self.autosign_file_path, self.autosign_file_permissions)

        self.run_call("test.ping -l quiet")  # get minion to authenticate itself again

        try:
            if os.path.isdir(self.autosign_grains_dir):
                shutil.rmtree(self.autosign_grains_dir)
        except AttributeError:
            pass

    @pytest.mark.slow_test
    def test_autosign_grains_accept(self):
        grain_file_path = os.path.join(self.autosign_grains_dir, "test_grain")
        with salt.utils.files.fopen(grain_file_path, "w") as f:
            f.write(salt.utils.stringutils.to_str("#invalid_value\ncheese"))
        os.chmod(grain_file_path, self.autosign_file_permissions)

        self.run_call(
            "test.ping -l quiet"
        )  # get minion to try to authenticate itself again
        self.assertIn("minion", self.run_key("-l acc"))

    @pytest.mark.slow_test
    def test_autosign_grains_fail(self):
        grain_file_path = os.path.join(self.autosign_grains_dir, "test_grain")
        with salt.utils.files.fopen(grain_file_path, "w") as f:
            f.write(salt.utils.stringutils.to_str("#cheese\ninvalid_value"))
        os.chmod(grain_file_path, self.autosign_file_permissions)

        self.run_call(
            "test.ping -l quiet"
        )  # get minion to try to authenticate itself again
        self.assertNotIn("minion", self.run_key("-l acc"))
        self.assertIn("minion", self.run_key("-l un"))



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import pytest
import salt.utils.platform
from tests.support.case import ShellCase


@pytest.mark.windows_whitelisted
@pytest.mark.usefixtures("salt_sub_minion")
class BatchTest(ShellCase):
    """
    Integration tests for the salt.cli.batch module
    """

    if salt.utils.platform.is_windows():
        run_timeout = 180
    else:
        run_timeout = 30

    @pytest.mark.slow_test
    def test_batch_run(self):
        """
        Tests executing a simple batch command to help catch regressions
        """
        ret = "Executing run on [{}]".format(repr("sub_minion"))
        cmd = self.run_salt(
            '"*minion" test.echo "batch testing" -b 50%',
            timeout=self.run_timeout,
        )
        self.assertIn(ret, cmd)

    @pytest.mark.slow_test
    def test_batch_run_number(self):
        """
        Tests executing a simple batch command using a number division instead of
        a percentage with full batch CLI call.
        """
        ret = "Executing run on [{}, {}]".format(repr("minion"), repr("sub_minion"))
        cmd = self.run_salt(
            '"*minion" test.ping --batch-size 2',
            timeout=self.run_timeout,
        )
        self.assertIn(ret, cmd)

    @pytest.mark.slow_test
    def test_batch_run_grains_targeting(self):
        """
        Tests executing a batch command using a percentage divisor as well as grains
        targeting.
        """
        os_grain = ""
        sub_min_ret = "Executing run on [{}]".format(repr("sub_minion"))
        min_ret = "Executing run on [{}]".format(repr("minion"))

        for item in self.run_salt("minion grains.get os"):
            if item != "minion:":
                os_grain = item

        os_grain = os_grain.strip()
        cmd = self.run_salt(
            '-C "G@os:{} and not localhost" -b 25% test.ping'.format(
                os_grain.replace(" ", "?")
            ),
            timeout=self.run_timeout,
        )
        self.assertIn(sub_min_ret, cmd)
        self.assertIn(min_ret, cmd)

    @pytest.mark.slow_test
    def test_batch_exit_code(self):
        """
        Test that a failed state returns a non-zero exit code in batch mode
        """
        cmd = self.run_salt(
            ' "*" state.single test.fail_without_changes name=test_me -b 25%',
            with_retcode=True,
            timeout=self.run_timeout,
        )
        self.assertEqual(cmd[-1], 2)

    # Test for failhard + batch. The best possible solution here was to do something like that:
    # assertRaises(StopIteration)
    # But it's impossible due to nature of the tests execution via fork()

    @pytest.mark.slow_test
    def test_batch_module_stopping_after_error(self):
        """
        Test that a failed command stops the batch run
        """

        minions_list = []
        retcode = None

        # Executing salt with batch: 1 and with failhard. It should stop after the first error.
        cmd = self.run_salt(
            '"*minion" test.retcode 42 -b 1 --out=yaml --failhard',
            timeout=self.run_timeout,
        )

        # Parsing the output. Idea is to fetch number on minions and retcode of the execution.
        # retcode var could be overwritten in case of broken failhard but number of minions check should still fail.
        for line in cmd:
            if line.startswith("Executing run on"):
                minions_list.append(line)
            if line.startswith("retcode"):
                retcode = line[-1]
        # We expect to have only one minion to be run
        self.assertEqual(1, len(minions_list))
        # We expect to find a retcode in the output
        self.assertIsNot(None, retcode)
        # We expect retcode to be non-zero
        self.assertNotEqual(0, retcode)

    @pytest.mark.slow_test
    def test_batch_state_stopping_after_error(self):
        """
        Test that a failed state stops the batch run
        """

        minions_list = []
        retcode = None

        # Executing salt with batch: 1 and with failhard. It should stop after the first error.
        cmd = self.run_salt(
            '"*minion" state.single test.fail_without_changes name=test_me -b 1'
            " --out=yaml --failhard",
            timeout=self.run_timeout,
        )

        # Parsing the output. Idea is to fetch number on minions and retcode of the execution.
        # retcode var could be overwritten in case of broken failhard but number of minions check should still fail.
        for line in cmd:
            if line.startswith("Executing run on"):
                minions_list.append(line)
            if line.startswith("retcode"):
                retcode = line[-1]
        # We expect to have only one minion to be run
        self.assertEqual(1, len(minions_list))
        # We expect to find a retcode in the output
        self.assertIsNot(None, retcode)
        # We expect retcode to be non-zero
        self.assertNotEqual(0, retcode)



"""
This module contains routines shared by the virt system.
"""

import hashlib
import logging
import os
import re
import time
import urllib
import urllib.parse

import salt.utils.files

# pylint: disable=E0611

log = logging.getLogger(__name__)


def download_remote(url, dir):
    """
    Attempts to download a file specified by 'url'

    :param url: The full remote path of the file which should be downloaded.
    :param dir: The path the file should be downloaded to.
    """

    try:
        rand = hashlib.md5(os.urandom(32)).hexdigest()
        remote_filename = urllib.parse.urlparse(url).path.split("/")[-1]
        full_directory = os.path.join(dir, "{}-{}".format(rand, remote_filename))
        with salt.utils.files.fopen(
            full_directory, "wb"
        ) as file, urllib.request.urlopen(url) as response:
            file.write(response.rease())

        return full_directory

    except Exception as err:  # pylint: disable=broad-except
        raise err


def check_remote(cmdline_path):
    """
    Checks to see if the path provided contains ftp, http, or https. Returns
    the full path if it is found.

    :param cmdline_path: The path to the initrd image or the kernel
    """
    regex = re.compile("^(ht|f)tps?\\b")

    if regex.match(urllib.parse.urlparse(cmdline_path).scheme):
        return True

    return False


class VirtKey:
    """
    Used to manage key signing requests.
    """

    def __init__(self, hyper, id_, opts):
        self.opts = opts
        self.hyper = hyper
        self.id = id_
        path = os.path.join(self.opts["pki_dir"], "virtkeys", hyper)
        if not os.path.isdir(path):
            os.makedirs(path)
        self.path = os.path.join(path, id_)

    def accept(self, pub):
        """
        Accept the provided key
        """
        try:
            with salt.utils.files.fopen(self.path, "r") as fp_:
                expiry = int(fp_.read())
        except OSError:
            log.error(
                "Request to sign key for minion '%s' on hyper '%s' "
                "denied: no authorization",
                self.id,
                self.hyper,
            )
            return False
        except ValueError:
            log.error("Invalid expiry data in %s", self.path)
            return False

        # Limit acceptance window to 10 minutes
        # TODO: Move this value to the master config file
        if (time.time() - expiry) > 600:
            log.warning(
                'Request to sign key for minion "%s" on hyper "%s" denied: '
                "authorization expired",
                self.id,
                self.hyper,
            )
            return False

        pubfn = os.path.join(self.opts["pki_dir"], "minions", self.id)
        with salt.utils.files.fopen(pubfn, "w+") as fp_:
            fp_.write(pub)
        self.void()
        return True

    def authorize(self):
        """
        Prepare the master to expect a signing request
        """
        with salt.utils.files.fopen(self.path, "w+") as fp_:
            fp_.write(str(int(time.time())))
        return True

    def void(self):
        """
        Invalidate any existing authorization
        """
        try:
            os.unlink(self.path)
            return True
        except OSError:
            return False



"""
The rcctl service module for OpenBSD
"""

import os

import salt.utils.decorators as decorators
import salt.utils.path
from salt.exceptions import CommandNotFoundError

__func_alias__ = {"reload_": "reload"}

# Define the module's virtual name
__virtualname__ = "service"


def __virtual__():
    """
    rcctl(8) is only available on OpenBSD.
    """
    if __grains__["os"] == "OpenBSD" and os.path.exists("/usr/sbin/rcctl"):
        return __virtualname__
    return (
        False,
        "The openbsdpkg execution module cannot be loaded: "
        "only available on OpenBSD systems.",
    )


@decorators.memoize
def _cmd():
    """
    Return the full path to the rcctl(8) command.
    """
    rcctl = salt.utils.path.which("rcctl")
    if not rcctl:
        raise CommandNotFoundError
    return rcctl


def _get_flags(**kwargs):
    """
    Return the configured service flags.
    """
    flags = kwargs.get("flags", __salt__["config.option"]("service.flags", default=""))
    return flags


def available(name):
    """
    Return True if the named service is available.

    CLI Example:

    .. code-block:: bash

        salt '*' service.available sshd
    """
    cmd = "{} get {}".format(_cmd(), name)
    if __salt__["cmd.retcode"](cmd) == 2:
        return False
    return True


def missing(name):
    """
    The inverse of service.available.
    Return True if the named service is not available.

    CLI Example:

    .. code-block:: bash

        salt '*' service.missing sshd
    """
    return not available(name)


def get_all():
    """
    Return all installed services.

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_all
    """
    ret = []
    service = _cmd()
    for svc in __salt__["cmd.run"]("{} ls all".format(service)).splitlines():
        ret.append(svc)
    return sorted(ret)


def get_disabled():
    """
    Return what services are available but not enabled to start at boot.

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_disabled
    """
    ret = []
    service = _cmd()
    for svc in __salt__["cmd.run"]("{} ls off".format(service)).splitlines():
        ret.append(svc)
    return sorted(ret)


def get_enabled():
    """
    Return what services are set to run on boot.

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_enabled
    """
    ret = []
    service = _cmd()
    for svc in __salt__["cmd.run"]("{} ls on".format(service)).splitlines():
        ret.append(svc)
    return sorted(ret)


def start(name):
    """
    Start the named service.

    CLI Example:

    .. code-block:: bash

        salt '*' service.start <service name>
    """
    cmd = "{} -f start {}".format(_cmd(), name)
    return not __salt__["cmd.retcode"](cmd)


def stop(name):
    """
    Stop the named service.

    CLI Example:

    .. code-block:: bash

        salt '*' service.stop <service name>
    """
    cmd = "{} stop {}".format(_cmd(), name)
    return not __salt__["cmd.retcode"](cmd)


def restart(name):
    """
    Restart the named service.

    CLI Example:

    .. code-block:: bash

        salt '*' service.restart <service name>
    """
    cmd = "{} -f restart {}".format(_cmd(), name)
    return not __salt__["cmd.retcode"](cmd)


def reload_(name):
    """
    Reload the named service.

    CLI Example:

    .. code-block:: bash

        salt '*' service.reload <service name>
    """
    cmd = "{} reload {}".format(_cmd(), name)
    return not __salt__["cmd.retcode"](cmd)


def status(name, sig=None):
    """
    Return the status for a service, returns a bool whether the service is
    running.

    CLI Example:

    .. code-block:: bash

        salt '*' service.status <service name>
    """
    if sig:
        return bool(__salt__["status.pid"](sig))

    cmd = "{} check {}".format(_cmd(), name)
    return not __salt__["cmd.retcode"](cmd)


def enable(name, **kwargs):
    """
    Enable the named service to start at boot.

    flags : None
        Set optional flags to run the service with.

    service.flags can be used to change the default flags.

    CLI Example:

    .. code-block:: bash

        salt '*' service.enable <service name>
        salt '*' service.enable <service name> flags=<flags>
    """
    stat_cmd = "{} set {} status on".format(_cmd(), name)
    stat_retcode = __salt__["cmd.retcode"](stat_cmd)

    flag_retcode = None
    # only (re)set flags for services that have an rc.d(8) script
    if os.path.exists("/etc/rc.d/{}".format(name)):
        flags = _get_flags(**kwargs)
        flag_cmd = "{} set {} flags {}".format(_cmd(), name, flags)
        flag_retcode = __salt__["cmd.retcode"](flag_cmd)

    return not any([stat_retcode, flag_retcode])


def disable(name, **kwargs):
    """
    Disable the named service to not start at boot.

    CLI Example:

    .. code-block:: bash

        salt '*' service.disable <service name>
    """
    cmd = "{} set {} status off".format(_cmd(), name)
    return not __salt__["cmd.retcode"](cmd)


def disabled(name):
    """
    Return True if the named service is disabled at boot, False otherwise.

    CLI Example:

    .. code-block:: bash

        salt '*' service.disabled <service name>
    """
    cmd = "{} get {} status".format(_cmd(), name)
    return not __salt__["cmd.retcode"](cmd) == 0


def enabled(name, **kwargs):
    """
    Return True if the named service is enabled at boot and the provided
    flags match the configured ones (if any). Return False otherwise.

    name
        Service name

    CLI Example:

    .. code-block:: bash

        salt '*' service.enabled <service name>
        salt '*' service.enabled <service name> flags=<flags>
    """
    cmd = "{} get {} status".format(_cmd(), name)
    if not __salt__["cmd.retcode"](cmd):
        # also consider a service disabled if the current flags are different
        # than the configured ones so we have a chance to update them
        flags = _get_flags(**kwargs)
        cur_flags = __salt__["cmd.run_stdout"]("{} get {} flags".format(_cmd(), name))
        if format(flags) == format(cur_flags):
            return True
        if not flags:
            def_flags = __salt__["cmd.run_stdout"](
                "{} getdef {} flags".format(_cmd(), name)
            )
            if format(cur_flags) == format(def_flags):
                return True

    return False



from tests.support.case import ModuleCase
from tests.support.jinja_filters import JinjaFiltersTest


class StateModuleJinjaFiltersTest(ModuleCase, JinjaFiltersTest):
    """
    testing Jinja filters are available via state system
    """



"""
XenServer Cloud Driver
======================

The XenServer driver is designed to work with a Citrix XenServer.

Requires XenServer SDK
(can be downloaded from https://www.citrix.com/downloads/xenserver/product-software/ )

Place a copy of the XenAPI.py in the Python site-packages folder.

:depends: XenAPI

Example provider configuration:

    .. code-block:: yaml

        # /etc/salt/cloud.providers.d/myxen.conf
        myxen:
          driver: xen
          url: http://10.0.0.120
          user: root
          password: p@ssw0rd

Example profile configuration:

    .. code-block:: yaml

        # /etc/salt/cloud.profiles.d/myxen.conf
        suse:
          provider: myxen
          user: root
          password: p@ssw0rd
          image: opensuseleap42_2-template
          storage_repo: 'Local storage'
          resource_pool: default_pool
          clone: True
          minion:
            master: 10.0.0.18
        sles:
          provider: myxen
          user: root
          clone: False
          image: sles12sp2-template
          deploy: False
        w2k12:
          provider: myxen
          image: w2k12svr-template
          clone: True
          userdata_file: /srv/salt/win/files/windows-firewall.ps1
          win_installer: /srv/salt/win/files/Salt-Minion-2016.11.3-AMD64-Setup.exe
          win_username: Administrator
          win_password: p@ssw0rd
          use_winrm: False
          ipv4_cidr: 10.0.0.215/24
          ipv4_gw: 10.0.0.1

"""

import logging
import time
from datetime import datetime

import salt.config as config
import salt.utils.cloud
from salt.exceptions import SaltCloudException, SaltCloudSystemExit

# Get logging started
log = logging.getLogger(__name__)

try:
    import XenAPI

    HAS_XEN_API = True
except ImportError:
    HAS_XEN_API = False

__virtualname__ = "xen"
cache = None


def __virtual__():
    """
    Only load if Xen configuration and XEN SDK is found.
    """
    if get_configured_provider() is False:
        return False
    if _get_dependencies() is False:
        return False

    global cache  # pylint: disable=global-statement,invalid-name
    cache = salt.cache.Cache(__opts__)

    return __virtualname__


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def _get_dependencies():
    """
    Warn if dependencies aren't met.

    Checks for the XenAPI.py module
    """
    return config.check_driver_dependencies(__virtualname__, {"XenAPI": HAS_XEN_API})


def get_configured_provider():
    """
    Return the first configured instance.
    """
    return config.is_provider_configured(
        __opts__, _get_active_provider_name() or __virtualname__, ("url",)
    )


def _get_session():
    """
    Get a connection to the XenServer host
    """
    api_version = "1.0"
    originator = "salt_cloud_{}_driver".format(__virtualname__)
    url = config.get_cloud_config_value(
        "url", get_configured_provider(), __opts__, search_global=False
    )
    user = config.get_cloud_config_value(
        "user", get_configured_provider(), __opts__, search_global=False
    )
    password = config.get_cloud_config_value(
        "password", get_configured_provider(), __opts__, search_global=False
    )
    ignore_ssl = config.get_cloud_config_value(
        "ignore_ssl",
        get_configured_provider(),
        __opts__,
        default=False,
        search_global=False,
    )
    try:
        session = XenAPI.Session(url, ignore_ssl=ignore_ssl)
        log.debug(
            "url: %s user: %s password: %s, originator: %s",
            url,
            user,
            "XXX-pw-redacted-XXX",
            originator,
        )
        session.xenapi.login_with_password(user, password, api_version, originator)
    except XenAPI.Failure as ex:
        pool_master_addr = str(ex.__dict__["details"][1])
        slash_parts = url.split("/")
        new_url = "/".join(slash_parts[:2]) + "/" + pool_master_addr
        session = XenAPI.Session(new_url)
        log.debug(
            "session is -> url: %s user: %s password: %s, originator:%s",
            new_url,
            user,
            "XXX-pw-redacted-XXX",
            originator,
        )
        session.xenapi.login_with_password(user, password, api_version, originator)
    return session


def list_nodes():
    """
    List virtual machines

      .. code-block:: bash

          salt-cloud -Q

    """
    session = _get_session()
    vms = session.xenapi.VM.get_all_records()
    ret = {}
    for vm in vms:
        record = session.xenapi.VM.get_record(vm)
        if not record["is_a_template"] and not record["is_control_domain"]:
            try:
                base_template_name = record["other_config"]["base_template_name"]
            except Exception:  # pylint: disable=broad-except
                base_template_name = None
                log.debug(
                    "VM %s, does not have base_template_name attribute",
                    record["name_label"],
                )
            ret[record["name_label"]] = {
                "id": record["uuid"],
                "image": base_template_name,
                "name": record["name_label"],
                "size": record["memory_dynamic_max"],
                "state": record["power_state"],
                "private_ips": get_vm_ip(record["name_label"], session),
                "public_ips": None,
            }
    return ret


def get_vm_ip(name=None, session=None, call=None):
    """
    Get the IP address of the VM

    .. code-block:: bash

        salt-cloud -a get_vm_ip xenvm01

    .. note:: Requires xen guest tools to be installed in VM

    """
    if call == "function":
        raise SaltCloudException("This function must be called with -a or --action.")
    if session is None:
        log.debug("New session being created")
        session = _get_session()
    vm = _get_vm(name, session=session)
    ret = None
    # -- try to get ip from vif
    vifs = session.xenapi.VM.get_VIFs(vm)
    if vifs is not None:
        for vif in vifs:
            if session.xenapi.VIF.get_ipv4_addresses(vif):
                cidr = session.xenapi.VIF.get_ipv4_addresses(vif).pop()
                ret, subnet = cidr.split("/")
                log.debug("VM vif returned for instance: %s ip: %s", name, ret)
                return ret
    # -- try to get ip from get tools metrics
    vgm = session.xenapi.VM.get_guest_metrics(vm)
    try:
        net = session.xenapi.VM_guest_metrics.get_networks(vgm)
        if "0/ip" in net.keys():
            log.debug(
                "VM guest metrics returned for instance: %s 0/ip: %s", name, net["0/ip"]
            )
            ret = net["0/ip"]
    # except Exception as ex:  # pylint: disable=broad-except
    except XenAPI.Failure:
        log.info("Could not get vm metrics at this time")
    return ret


def set_vm_ip(name=None, ipv4_cidr=None, ipv4_gw=None, session=None, call=None):
    """
    Set the IP address on a virtual interface (vif)

    """
    mode = "static"
    # TODO: Need to add support for IPv6
    if call == "function":
        raise SaltCloudException("The function must be called with -a or --action.")

    log.debug(
        "Setting name: %s ipv4_cidr: %s ipv4_gw: %s mode: %s",
        name,
        ipv4_cidr,
        ipv4_gw,
        mode,
    )
    if session is None:
        log.debug("New session being created")
        session = _get_session()
    vm = _get_vm(name, session)
    # -- try to get ip from vif
    # TODO: for now will take first interface
    #       addition consideration needed for
    #       multiple interface(vif) VMs
    vifs = session.xenapi.VM.get_VIFs(vm)
    if vifs is not None:
        log.debug("There are %s vifs.", len(vifs))
        for vif in vifs:
            record = session.xenapi.VIF.get_record(vif)
            log.debug(record)
            try:
                session.xenapi.VIF.configure_ipv4(vif, mode, ipv4_cidr, ipv4_gw)
            except XenAPI.Failure:
                log.info("Static IP assignment could not be performed.")

    return True


def list_nodes_full(session=None):
    """
    List full virtual machines

      .. code-block:: bash

          salt-cloud -F

    """
    if session is None:
        session = _get_session()

    ret = {}
    vms = session.xenapi.VM.get_all()
    for vm in vms:
        record = session.xenapi.VM.get_record(vm)
        if not record["is_a_template"] and not record["is_control_domain"]:
            # deal with cases where the VM doesn't have 'base_template_name' attribute
            try:
                base_template_name = record["other_config"]["base_template_name"]
            except Exception:  # pylint: disable=broad-except
                base_template_name = None
                log.debug(
                    "VM %s, does not have base_template_name attribute",
                    record["name_label"],
                )
            vm_cfg = session.xenapi.VM.get_record(vm)
            vm_cfg["id"] = record["uuid"]
            vm_cfg["name"] = record["name_label"]
            vm_cfg["image"] = base_template_name
            vm_cfg["size"] = None
            vm_cfg["state"] = record["power_state"]
            vm_cfg["private_ips"] = get_vm_ip(record["name_label"], session)
            vm_cfg["public_ips"] = None
            if "snapshot_time" in vm_cfg.keys():
                del vm_cfg["snapshot_time"]
            ret[record["name_label"]] = vm_cfg

    provider = _get_active_provider_name() or "xen"
    if ":" in provider:
        comps = provider.split(":")
        provider = comps[0]
    log.debug("ret: %s", ret)
    log.debug("provider: %s", provider)
    log.debug("__opts__: %s", __opts__)
    __utils__["cloud.cache_node_list"](ret, provider, __opts__)
    return ret


def list_nodes_select(call=None):
    """
    Perform a select query on Xen VM instances

    .. code-block:: bash

        salt-cloud -S

    """
    return salt.utils.cloud.list_nodes_select(
        list_nodes_full(),
        __opts__["query.selection"],
        call,
    )


def vdi_list(call=None, kwargs=None):
    """
    Return available Xen VDI images

    If this function is called with the ``-f`` or ``--function`` then
    it can return a list with minimal deatil using the ``terse=True`` keyword
    argument.

    .. code-block:: bash

        salt-cloud -f vdi_list myxen terse=True

    """
    if call == "action":
        raise SaltCloudException("This function must be called with -f or --function.")
    log.debug("kwargs is %s", kwargs)
    if kwargs is not None:
        if "terse" in kwargs:
            if kwargs["terse"] == "True":
                terse = True
            else:
                terse = False
        else:
            terse = False
    else:
        kwargs = {}
        terse = False
    session = _get_session()
    vdis = session.xenapi.VDI.get_all()
    ret = {}
    for vdi in vdis:
        data = session.xenapi.VDI.get_record(vdi)
        log.debug(type(terse))
        if terse is True:
            ret[data.get("name_label")] = {"uuid": data.get("uuid"), "OpqueRef": vdi}
        else:
            data.update({"OpaqueRef": vdi})
            ret[data.get("name_label")] = data
    return ret


def avail_locations(session=None, call=None):
    """
    Return available Xen locations (not implemented)

    .. code-block:: bash

        salt-cloud --list-locations myxen

    """
    # TODO: need to figure out a good meaning of locations in Xen
    if call == "action":
        raise SaltCloudException(
            "The avail_locations function must be called with -f or --function."
        )
    return pool_list()


def avail_sizes(session=None, call=None):
    """
    Return a list of Xen template definitions

    .. code-block:: bash

        salt-cloud --list-sizes myxen

    """
    if call == "action":
        raise SaltCloudException(
            "The avail_sizes function must be called with -f or --function."
        )
    return {
        "STATUS": (
            "Sizes are build into templates. Consider running --list-images to see"
            " sizes"
        )
    }


def template_list(call=None):
    """
    Return available Xen template information.

    This returns the details of
    each template to show number cores, memory sizes, etc..

    .. code-block:: bash

       salt-cloud -f template_list myxen

    """
    templates = {}
    session = _get_session()
    vms = session.xenapi.VM.get_all()
    for vm in vms:
        record = session.xenapi.VM.get_record(vm)
        if record["is_a_template"]:
            templates[record["name_label"]] = record
    return templates


def show_instance(name, session=None, call=None):
    """
    Show information about a specific VM or template

        .. code-block:: bash

            salt-cloud -a show_instance xenvm01

    .. note:: memory is memory_dynamic_max

    """
    if call == "function":
        raise SaltCloudException(
            "The show_instnce function must be called with -a or --action."
        )
    log.debug("show_instance-> name: %s session: %s", name, session)
    if session is None:
        session = _get_session()
    vm = _get_vm(name, session=session)
    record = session.xenapi.VM.get_record(vm)
    if not record["is_a_template"] and not record["is_control_domain"]:
        try:
            base_template_name = record["other_config"]["base_template_name"]
        except Exception:  # pylint: disable=broad-except
            base_template_name = None
            log.debug(
                "VM %s, does not have base_template_name attribute",
                record["name_label"],
            )
        ret = {
            "id": record["uuid"],
            "image": base_template_name,
            "name": record["name_label"],
            "size": record["memory_dynamic_max"],
            "state": record["power_state"],
            "private_ips": get_vm_ip(name, session),
            "public_ips": None,
        }

        __utils__["cloud.cache_node"](ret, _get_active_provider_name(), __opts__)
    return ret


def _determine_resource_pool(session, vm_):
    """
    Called by create() used to determine resource pool
    """
    resource_pool = ""
    if "resource_pool" in vm_.keys():
        resource_pool = _get_pool(vm_["resource_pool"], session)
    else:
        pool = session.xenapi.pool.get_all()
        if not pool:
            resource_pool = None
        else:
            first_pool = session.xenapi.pool.get_all()[0]
            resource_pool = first_pool
    pool_record = session.xenapi.pool.get_record(resource_pool)
    log.debug("resource pool: %s", pool_record["name_label"])
    return resource_pool


def _determine_storage_repo(session, resource_pool, vm_):
    """
    Called by create() used to determine storage repo for create
    """
    storage_repo = ""
    if "storage_repo" in vm_.keys():
        storage_repo = _get_sr(vm_["storage_repo"], session)
    else:
        storage_repo = None
        if resource_pool:
            default_sr = session.xenapi.pool.get_default_SR(resource_pool)
            sr_record = session.xenapi.SR.get_record(default_sr)
            log.debug("storage repository: %s", sr_record["name_label"])
            storage_repo = default_sr
        else:
            storage_repo = None
    log.debug("storage repository: %s", storage_repo)
    return storage_repo


def create(vm_):
    """
    Create a VM in Xen

    The configuration for this function is read from the profile settings.

    .. code-block:: bash

        salt-cloud -p some_profile xenvm01

    """
    name = vm_["name"]
    record = {}
    ret = {}

    # fire creating event
    __utils__["cloud.fire_event"](
        "event",
        "starting create",
        "salt/cloud/{}/creating".format(name),
        args={"name": name, "profile": vm_["profile"], "provider": vm_["driver"]},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )
    log.debug("Adding %s to cloud cache.", name)
    __utils__["cloud.cachedir_index_add"](
        vm_["name"], vm_["profile"], "xen", vm_["driver"]
    )

    # connect to xen
    session = _get_session()

    # determine resource pool
    resource_pool = _determine_resource_pool(session, vm_)

    # determine storage repo
    storage_repo = _determine_storage_repo(session, resource_pool, vm_)

    # build VM
    image = vm_.get("image")
    clone = vm_.get("clone")
    if clone is None:
        clone = True
    log.debug("Clone: %s ", clone)

    # fire event to read new vm properties (requesting)
    __utils__["cloud.fire_event"](
        "event",
        "requesting instance",
        "salt/cloud/{}/requesting".format(name),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    # create by cloning template
    if clone:
        _clone_vm(image, name, session)
    else:
        _copy_vm(image, name, session, storage_repo)

    # provision template to vm
    _provision_vm(name, session)
    vm = _get_vm(name, session)

    # start vm
    start(name, None, session)

    # get new VM
    vm = _get_vm(name, session)

    # wait for vm to report IP via guest tools
    _wait_for_ip(name, session)

    # set static IP if configured
    _set_static_ip(name, session, vm_)

    # if not deploying salt then exit
    deploy = vm_.get("deploy", True)
    log.debug("delopy is set to %s", deploy)
    if deploy:
        record = session.xenapi.VM.get_record(vm)
        if record is not None:
            _deploy_salt_minion(name, session, vm_)
    else:
        log.debug("The Salt minion will not be installed, deploy: %s", vm_["deploy"])
    record = session.xenapi.VM.get_record(vm)
    ret = show_instance(name)
    ret.update({"extra": record})

    __utils__["cloud.fire_event"](
        "event",
        "created instance",
        "salt/cloud/{}/created".format(name),
        args={"name": name, "profile": vm_["profile"], "provider": vm_["driver"]},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )
    return ret


def _deploy_salt_minion(name, session, vm_):
    """
    Deploy salt minion during create()
    """
    # Get bootstrap values
    vm_["ssh_host"] = get_vm_ip(name, session)
    vm_["user"] = vm_.get("user", "root")
    vm_["password"] = vm_.get("password", "p@ssw0rd!")
    vm_["provider"] = vm_.get("provider", "xen")
    log.debug("%s has IP of %s", name, vm_["ssh_host"])
    # Bootstrap Salt minion!
    if vm_["ssh_host"] is not None:
        log.info("Installing Salt minion on %s", name)
        boot_ret = __utils__["cloud.bootstrap"](vm_, __opts__)
        log.debug("boot return: %s", boot_ret)


def _set_static_ip(name, session, vm_):
    """
    Set static IP during create() if defined
    """
    ipv4_cidr = ""
    ipv4_gw = ""
    if "ipv4_gw" in vm_.keys():
        log.debug("ipv4_gw is found in keys")
        ipv4_gw = vm_["ipv4_gw"]
    if "ipv4_cidr" in vm_.keys():
        log.debug("ipv4_cidr is found in keys")
        ipv4_cidr = vm_["ipv4_cidr"]
        log.debug("attempting to set IP in instance")
        set_vm_ip(name, ipv4_cidr, ipv4_gw, session, None)


def _wait_for_ip(name, session):
    """
    Wait for IP  to be available during create()
    """
    start_time = datetime.now()
    status = None
    while status is None:
        status = get_vm_ip(name, session)
        if status is not None:
            # ignore APIPA address
            if status.startswith("169"):
                status = None
        check_time = datetime.now()
        delta = check_time - start_time
        log.debug(
            "Waited %s seconds for %s to report ip address...", delta.seconds, name
        )
        if delta.seconds > 180:
            log.warning("Timeout getting IP address")
            break
        time.sleep(5)


def _run_async_task(task=None, session=None):
    """
    Run  XenAPI task in asynchronous mode to prevent timeouts
    """
    if task is None or session is None:
        return None
    task_name = session.xenapi.task.get_name_label(task)
    log.debug("Running %s", task_name)
    while session.xenapi.task.get_status(task) == "pending":
        progress = round(session.xenapi.task.get_progress(task), 2) * 100
        log.debug("Task progress %.2f%%", progress)
        time.sleep(1)
    log.debug("Cleaning up task %s", task_name)
    session.xenapi.task.destroy(task)


def _clone_vm(image=None, name=None, session=None):
    """
    Create VM by cloning

    This is faster and should be used if source and target are
    in the same storage repository

    """
    if session is None:
        session = _get_session()
    log.debug("Creating VM %s by cloning %s", name, image)
    source = _get_vm(image, session)
    task = session.xenapi.Async.VM.clone(source, name)
    _run_async_task(task, session)


def _copy_vm(template=None, name=None, session=None, sr=None):
    """
    Create VM by copy

    This is slower and should be used if source and target are
    NOT in the same storage repository

    template = object reference
    name = string name of new VM
    session = object reference
    sr = object reference
    """
    if session is None:
        session = _get_session()
    log.debug("Creating VM %s by copying %s", name, template)
    source = _get_vm(template, session)
    task = session.xenapi.Async.VM.copy(source, name, sr)
    _run_async_task(task, session)


def _provision_vm(name=None, session=None):
    """
    Provision vm right after clone/copy
    """
    if session is None:
        session = _get_session()
    log.info("Provisioning VM %s", name)
    vm = _get_vm(name, session)
    task = session.xenapi.Async.VM.provision(vm)
    _run_async_task(task, session)


def start(name, call=None, session=None):
    """
    Start  a vm

    .. code-block:: bash

        salt-cloud -a start xenvm01

    """
    if call == "function":
        raise SaltCloudException(
            "The show_instnce function must be called with -a or --action."
        )
    if session is None:
        session = _get_session()
    log.info("Starting VM %s", name)
    vm = _get_vm(name, session)
    task = session.xenapi.Async.VM.start(vm, False, True)
    _run_async_task(task, session)
    return show_instance(name)


def pause(name, call=None, session=None):
    """
    Pause a vm

    .. code-block:: bash

        salt-cloud -a pause xenvm01

    """
    if call == "function":
        raise SaltCloudException(
            "The show_instnce function must be called with -a or --action."
        )
    if session is None:
        session = _get_session()
    log.info("Pausing VM %s", name)
    vm = _get_vm(name, session)
    task = session.xenapi.Async.VM.pause(vm)
    _run_async_task(task, session)
    return show_instance(name)


def unpause(name, call=None, session=None):
    """
    UnPause a vm

    .. code-block:: bash

        salt-cloud -a unpause xenvm01

    """
    if call == "function":
        raise SaltCloudException(
            "The show_instnce function must be called with -a or --action."
        )
    if session is None:
        session = _get_session()
    log.info("Unpausing VM %s", name)
    vm = _get_vm(name, session)
    task = session.xenapi.Async.VM.unpause(vm)
    _run_async_task(task, session)
    return show_instance(name)


def suspend(name, call=None, session=None):
    """
    Suspend a vm to disk

    .. code-block:: bash

        salt-cloud -a suspend xenvm01

    """
    if call == "function":
        raise SaltCloudException(
            "The show_instnce function must be called with -a or --action."
        )
    if session is None:
        session = _get_session()
    log.info("Suspending VM %s", name)
    vm = _get_vm(name, session)
    task = session.xenapi.Async.VM.suspend(vm)
    _run_async_task(task, session)
    return show_instance(name)


def resume(name, call=None, session=None):
    """
    Resume a vm from disk

    .. code-block:: bash

        salt-cloud -a resume xenvm01

    """
    if call == "function":
        raise SaltCloudException(
            "The show_instnce function must be called with -a or --action."
        )
    if session is None:
        session = _get_session()
    log.info("Resuming VM %s", name)
    vm = _get_vm(name, session)
    task = session.xenapi.Async.VM.resume(vm, False, True)
    _run_async_task(task, session)
    return show_instance(name)


def stop(name, call=None, session=None):
    """
    Stop a vm

    .. code-block:: bash

        salt-cloud -a stop xenvm01


    """
    if call == "function":
        raise SaltCloudException(
            "The show_instnce function must be called with -a or --action."
        )
    return shutdown(name, call, session)


def shutdown(name, call=None, session=None):
    """
    Shutdown  a vm

    .. code-block:: bash

        salt-cloud -a shutdown xenvm01

    """
    if call == "function":
        raise SaltCloudException(
            "The show_instnce function must be called with -a or --action."
        )
    if session is None:
        session = _get_session()
    log.info("Starting VM %s", name)
    vm = _get_vm(name, session)
    task = session.xenapi.Async.VM.shutdown(vm)
    _run_async_task(task, session)
    return show_instance(name)


def reboot(name, call=None, session=None):
    """
    Reboot a vm

    .. code-block:: bash

        salt-cloud -a reboot xenvm01

    """
    if call == "function":
        raise SaltCloudException(
            "The show_instnce function must be called with -a or --action."
        )
    if session is None:
        session = _get_session()
    log.info("Starting VM %s", name)
    vm = _get_vm(name, session)
    power_state = session.xenapi.VM.get_power_state(vm)
    if power_state == "Running":
        task = session.xenapi.Async.VM.clean_reboot(vm)
        _run_async_task(task, session)
        return show_instance(name)
    else:
        return "{} is not running to be rebooted".format(name)


def _get_vm(name=None, session=None):
    """
    Get XEN vm instance object reference
    """
    if session is None:
        session = _get_session()
    vms = session.xenapi.VM.get_by_name_label(name)
    vms = [x for x in vms if not session.xenapi.VM.get_is_a_template(x)]
    if len(vms) == 1:
        return vms[0]
    else:
        log.error("VM %s returned %s matches. 1 match expected.", name, len(vms))
    return None


def _get_sr(name=None, session=None):
    """
    Get XEN sr (storage repo) object reference
    """
    if session is None:
        session = _get_session()
    srs = session.xenapi.SR.get_by_name_label(name)
    if len(srs) == 1:
        return srs[0]
    return None


def _get_pool(name=None, session=None):
    """
    Get XEN resource pool object reference
    """
    if session is None:
        session = _get_session()
    pools = session.xenapi.pool.get_all()
    for pool in pools:
        pool_record = session.xenapi.pool.get_record(pool)
        if name in pool_record.get("name_label"):
            return pool
    return None


def destroy(name=None, call=None):
    """
    Destroy Xen VM or template instance

    .. code-block:: bash

        salt-cloud -d xenvm01

    """
    if call == "function":
        raise SaltCloudSystemExit(
            "The destroy action must be called with -d, --destroy, -a or --action."
        )
    ret = {}
    __utils__["cloud.fire_event"](
        "event",
        "destroying instance",
        "salt/cloud/{}/destroying".format(name),
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )
    session = _get_session()
    vm = _get_vm(name)
    if vm:
        # get vm
        record = session.xenapi.VM.get_record(vm)
        log.debug("power_state: %s", record["power_state"])
        # shut down
        if record["power_state"] != "Halted":
            task = session.xenapi.Async.VM.hard_shutdown(vm)
            _run_async_task(task, session)

        # destroy disk (vdi) by reading vdb on vm
        ret["vbd"] = destroy_vm_vdis(name, session)
        # destroy vm
        task = session.xenapi.Async.VM.destroy(vm)
        _run_async_task(task, session)
        ret["destroyed"] = True
        __utils__["cloud.fire_event"](
            "event",
            "destroyed instance",
            "salt/cloud/{}/destroyed".format(name),
            args={"name": name},
            sock_dir=__opts__["sock_dir"],
            transport=__opts__["transport"],
        )
        if __opts__.get("update_cachedir", False) is True:
            __utils__["cloud.delete_minion_cachedir"](
                name, _get_active_provider_name().split(":")[0], __opts__
            )
        __utils__["cloud.cachedir_index_del"](name)
        return ret


def sr_list(call=None):
    """
    Geta list of storage repositories

    .. code-block:: bash

        salt-cloud -f sr_list myxen

    """
    if call != "function":
        raise SaltCloudSystemExit(
            "This function must be called with -f, --function argument."
        )
    ret = {}
    session = _get_session()
    srs = session.xenapi.SR.get_all()
    for sr in srs:
        sr_record = session.xenapi.SR.get_record(sr)
        ret[sr_record["name_label"]] = sr_record
    return ret


def host_list(call=None):
    """
    Get a list of Xen Servers

    .. code-block:: bash

        salt-cloud -f host_list myxen
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "This function must be called with -f, --function argument."
        )
    ret = {}
    session = _get_session()
    hosts = session.xenapi.host.get_all()
    for host in hosts:
        host_record = session.xenapi.host.get_record(host)
        ret[host_record["name_label"]] = host_record
    return ret


def pool_list(call=None):
    """
    Get a list of Resource Pools

    .. code-block:: bash

        salt-cloud -f pool_list myxen

    """
    if call == "action":
        raise SaltCloudSystemExit(
            "This function must be called with -f, --function argument."
        )
    ret = {}
    session = _get_session()
    pools = session.xenapi.pool.get_all()
    for pool in pools:
        pool_record = session.xenapi.pool.get_record(pool)
        ret[pool_record["name_label"]] = pool_record
    return ret


def pif_list(call=None):
    """
    Get a list of Resource Pools

    .. code-block:: bash

        salt-cloud -f pool_list myxen
    """
    if call != "function":
        raise SaltCloudSystemExit(
            "This function must be called with -f, --function argument."
        )
    ret = {}
    session = _get_session()
    pifs = session.xenapi.PIF.get_all()
    for pif in pifs:
        record = session.xenapi.PIF.get_record(pif)
        ret[record["uuid"]] = record
    return ret


def vif_list(name, call=None, kwargs=None):
    """
    Get a list of virtual network interfaces  on a VM

    **requires**: the name of the vm with the vbd definition

    .. code-block:: bash

        salt-cloud -a vif_list xenvm01

    """
    if call == "function":
        raise SaltCloudSystemExit(
            "This function must be called with -a, --action argument."
        )
    if name is None:
        return "A name kwarg is rquired"
    ret = {}
    data = {}
    session = _get_session()
    vm = _get_vm(name)
    vifs = session.xenapi.VM.get_VIFs(vm)
    if vifs is not None:
        x = 0
        for vif in vifs:
            vif_record = session.xenapi.VIF.get_record(vif)
            data["vif-{}".format(x)] = vif_record
            x += 1
    ret[name] = data
    return ret


def vbd_list(name=None, call=None):
    """
    Get a list of VBDs on a VM

    **requires**: the name of the vm with the vbd definition

    .. code-block:: bash

        salt-cloud -a vbd_list xenvm01

    """
    if call == "function":
        raise SaltCloudSystemExit(
            "This function must be called with -a, --action argument."
        )
    if name is None:
        return "A name kwarg is rquired"
    ret = {}
    data = {}
    session = _get_session()
    vms = session.xenapi.VM.get_by_name_label(name)
    if len(vms) == 1:
        vm = vms[0]
        vbds = session.xenapi.VM.get_VBDs(vm)
        if vbds is not None:
            x = 0
            for vbd in vbds:
                vbd_record = session.xenapi.VBD.get_record(vbd)
                data["vbd-{}".format(x)] = vbd_record
                x += 1
    ret = data
    return ret


def avail_images(call=None):
    """
    Get a list of images from Xen

    If called with the `--list-images` then it returns
    images with all details.

    .. code-block:: bash

        salt-cloud --list-images myxen

    """
    if call == "action":
        raise SaltCloudSystemExit(
            "This function must be called with -f, --function argument."
        )
    return template_list()


def destroy_vm_vdis(name=None, session=None, call=None):
    """
    Get virtual block devices on VM

    .. code-block:: bash

        salt-cloud -a destroy_vm_vdis  xenvm01

    """
    if session is None:
        session = _get_session()
    ret = {}
    # get vm object
    vms = session.xenapi.VM.get_by_name_label(name)
    if len(vms) == 1:
        # read virtual block device (vdb)
        vbds = session.xenapi.VM.get_VBDs(vms[0])
        if vbds is not None:
            x = 0
            for vbd in vbds:
                vbd_record = session.xenapi.VBD.get_record(vbd)
                if vbd_record["VDI"] != "OpaqueRef:NULL":
                    # read vdi on vdb
                    vdi_record = session.xenapi.VDI.get_record(vbd_record["VDI"])
                    if "iso" not in vdi_record["name_label"]:
                        session.xenapi.VDI.destroy(vbd_record["VDI"])
                        ret["vdi-{}".format(x)] = vdi_record["name_label"]
                x += 1
    return ret


def destroy_template(name=None, call=None, kwargs=None):
    """
    Destroy Xen VM or template instance

        .. code-block:: bash

            salt-cloud -f destroy_template myxen name=testvm2

    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The destroy_template function must be called with  -f."
        )
    if kwargs is None:
        kwargs = {}
    name = kwargs.get("name", None)
    session = _get_session()
    vms = session.xenapi.VM.get_all_records()
    ret = {}
    found = False
    for vm in vms:
        record = session.xenapi.VM.get_record(vm)
        if record["is_a_template"]:
            if record["name_label"] == name:
                found = True
                # log.debug(record['name_label'])
                session.xenapi.VM.destroy(vm)
                ret[name] = {"status": "destroyed"}
    if not found:
        ret[name] = {"status": "not found"}
    return ret


def get_pv_args(name, session=None, call=None):
    """
    Get PV arguments for a VM

    .. code-block:: bash

        salt-cloud -a get_pv_args xenvm01

    """
    if call == "function":
        raise SaltCloudException("This function must be called with -a or --action.")
    if session is None:
        log.debug("New session being created")
        session = _get_session()
    vm = _get_vm(name, session=session)
    pv_args = session.xenapi.VM.get_PV_args(vm)
    if pv_args:
        return pv_args
    return None


def set_pv_args(name, kwargs=None, session=None, call=None):
    """
    Set PV arguments for a VM

    .. code-block:: bash

        salt-cloud -a set_pv_args xenvm01 pv_args="utf-8 graphical"

    """
    if call == "function":
        raise SaltCloudException("This function must be called with -a or --action.")
    if session is None:
        log.debug("New session being created")
        session = _get_session()
    vm = _get_vm(name, session=session)
    try:
        log.debug("Setting PV Args: %s", kwargs["pv_args"])
        session.xenapi.VM.set_PV_args(vm, str(kwargs["pv_args"]))
    except KeyError:
        log.error("No pv_args parameter found.")
        return False
    except XenAPI.Failure:
        log.info("Setting PV Args failed.")
        return False
    return True



import pytest
import salt.utils.platform
from tests.support.case import ModuleCase


@pytest.mark.windows_whitelisted
class StdTest(ModuleCase):
    """
    Test standard client calls
    """

    def setUp(self):
        self.TIMEOUT = 600 if salt.utils.platform.is_windows() else 10

    @pytest.mark.slow_test
    def test_cli(self):
        """
        Test cli function
        """
        cmd_iter = self.client.cmd_cli(
            "minion", "test.arg", ["foo", "bar", "baz"], kwarg={"qux": "quux"}
        )
        for ret in cmd_iter:
            data = ret["minion"]["ret"]
            self.assertEqual(data["args"], ["foo", "bar", "baz"])
            self.assertEqual(data["kwargs"]["qux"], "quux")

    @pytest.mark.slow_test
    def test_iter(self):
        """
        test cmd_iter
        """
        cmd_iter = self.client.cmd_iter(
            "minion", "test.arg", ["foo", "bar", "baz"], kwarg={"qux": "quux"}
        )
        for ret in cmd_iter:
            data = ret["minion"]["ret"]
            self.assertEqual(data["args"], ["foo", "bar", "baz"])
            self.assertEqual(data["kwargs"]["qux"], "quux")

    @pytest.mark.slow_test
    def test_iter_no_block(self):
        """
        test cmd_iter_no_block
        """
        cmd_iter = self.client.cmd_iter_no_block(
            "minion", "test.arg", ["foo", "bar", "baz"], kwarg={"qux": "quux"}
        )
        for ret in cmd_iter:
            if ret is None:
                continue
            data = ret["minion"]["ret"]
            self.assertEqual(data["args"], ["foo", "bar", "baz"])
            self.assertEqual(data["kwargs"]["qux"], "quux")

    @pytest.mark.slow_test
    def test_full_returns(self):
        """
        test cmd_iter
        """
        ret = self.client.cmd_full_return(
            "minion",
            "test.arg",
            ["foo", "bar", "baz"],
            timeout=self.TIMEOUT,
            kwarg={"qux": "quux"},
        )
        data = ret["minion"]["ret"]
        self.assertEqual(data["args"], ["foo", "bar", "baz"])
        self.assertEqual(data["kwargs"]["qux"], "quux")

    @pytest.mark.slow_test
    def test_kwarg_type(self):
        """
        Test that kwargs end up on the client as the same type
        """
        terrible_yaml_string = 'foo: ""\n# \''
        ret = self.client.cmd_full_return(
            "minion",
            "test.arg_type",
            ["a", 1],
            kwarg={"outer": {"a": terrible_yaml_string}, "inner": "value"},
            timeout=self.TIMEOUT,
        )
        data = ret["minion"]["ret"]
        self.assertIn(str.__name__, data["args"][0])
        self.assertIn("int", data["args"][1])
        self.assertIn("dict", data["kwargs"]["outer"])
        self.assertIn(str.__name__, data["kwargs"]["inner"])

    @pytest.mark.slow_test
    def test_full_return_kwarg(self):
        ret = self.client.cmd(
            "minion",
            "test.ping",
            full_return=True,
            timeout=self.TIMEOUT,
        )
        for mid, data in ret.items():
            self.assertIn("retcode", data)

    @pytest.mark.slow_test
    def test_cmd_arg_kwarg_parsing(self):
        ret = self.client.cmd(
            "minion",
            "test.arg_clean",
            arg=["foo", "bar=off", "baz={qux: 123}"],
            kwarg={"quux": "Quux"},
            timeout=self.TIMEOUT,
        )
        self.assertEqual(
            ret["minion"],
            {
                "args": ["foo"],
                "kwargs": {"bar": False, "baz": {"qux": 123}, "quux": "Quux"},
            },
        )



import json
import logging

import pytest
from tests.support.helpers import SaltVirtualEnv
from tests.support.pytest.helpers import FakeSaltExtension

pytestmark = [
    # These are slow because they create a virtualenv and install salt in it
    pytest.mark.slow_test,
]

log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def salt_extension(tmp_path_factory):
    with FakeSaltExtension(
        tmp_path_factory=tmp_path_factory, name="salt-ext-version-test"
    ) as extension:
        yield extension


def test_salt_extensions_in_versions_report(tmp_path, salt_extension):
    with SaltVirtualEnv(venv_dir=tmp_path / ".venv") as venv:
        # Install our extension into the virtualenv
        venv.install(str(salt_extension.srcdir))
        installed_packages = venv.get_installed_packages()
        assert salt_extension.name in installed_packages
        ret = venv.run_code(
            """
            import json
            import salt.version

            print(json.dumps(salt.version.versions_information()))
            """
        )
    versions_information = json.loads(ret.stdout)
    assert "Salt Extensions" in versions_information
    assert salt_extension.name in versions_information["Salt Extensions"]


def test_salt_extensions_absent_in_versions_report(tmp_path, salt_extension):
    """
    Ensure that the 'Salt Extensions' header does not show up when no extension is installed
    """
    with SaltVirtualEnv(venv_dir=tmp_path / ".venv") as venv:
        installed_packages = venv.get_installed_packages()
        assert salt_extension.name not in installed_packages
        ret = venv.run_code(
            """
            import json
            import salt.version

            print(json.dumps(salt.version.versions_information()))
            """
        )
    versions_information = json.loads(ret.stdout)
    assert "Salt Extensions" not in versions_information
